"""Facility for general mathematic utilities."""

from math import hypot

from ..config import SCREEN_WIDTH, SCREEN_HEIGHT


# XXX
# use the divmod approach here;
# also, maybe having the order of
# parameters reversed would be more intuitive?
def get_reaching_multiple(step_length, total_distance):
    """Return first multiple to reach or surpass distance.

    step_length
        An int describing the length of the step.
    total_distance
        An int describing the total distance to cover.

    This functions is very useful for simple scrolling
    features and was first implemented on the load game
    menu (in the main menu).

    >>> get_reaching_multiple(2, 11)
    12
    >>> get_reaching_multiple(10, 101)
    110
    >>> get_reaching_multiple(10, 100)
    100
    >>> get_reaching_multiple(10, 10)
    10
    >>> get_reaching_multiple(10, 9)
    10
    """
    if step_length >= total_distance:
        return step_length
    else:
        steps_no = total_distance // step_length
        rest = total_distance % step_length
        if rest:
            return step_length * (steps_no + 1)
        else:
            return step_length * steps_no


def calculate_jump(coordinates):
    """Return how much scroll is needed to reach coordinates.

    Reaching coordinates means scrolling screen horizontally
    by its entire width and/or vertically by its entire
    height until the coordinates are placed on the
    screen.

    calculate_jump -> (x, y)

    x and y are integers.

    coordinates
        Reference point towards which to scroll.

    >>> point = ( 640,  360)
    >>> calculate_jump(point)
    (0, 0)
    >>> point = (1450,  360)
    >>> calculate_jump(point)
    (-1280, 0)
    >>> point = (-640,  360)
    >>> calculate_jump(point)
    (1280, 0)
    >>> point = ( 640, -360)
    >>> calculate_jump(point)
    (0, 720)
    >>> point = ( 640, 1360)
    >>> calculate_jump(point)
    (0, -720)
    """
    x, y = coordinates

    horizontal_jumps = x // SCREEN_WIDTH
    vertical_jumps = y // SCREEN_HEIGHT

    horizontal_scroll = horizontal_jumps * SCREEN_WIDTH
    vertical_scroll = vertical_jumps * SCREEN_HEIGHT

    return (-horizontal_scroll, -vertical_scroll)


def unscroll_coordinates(coordinates):
    """Return unscrolled coordinates.

    The unscrolled coordinates are the coordinates a point
    in the game world would have if it were scrolled to the
    current screen keeping the same distance from the edges
    of the screen. That is, in the diagram below, it would
    be the coordinates of A if it were scrolled back to
    the current screen (as if it were the point B, a point
    which has the same distance of the edges of the screen
    that point A has)).

    current screen (player sees this region)
        |
        |             another neighbouring screen
        |             |   (player can't see this region)
        |             |
        v             v
    __________________________
    |            |            |
    |   . B      |   . A      |
    |            |            |
    |            |            |
    |____________|____________|

    unscroll_coordinates(coordinates) -> point

    coordinates
        Just a point consisting of a list or tuple with
        two integer coordinates.
    """
    abs_x, abs_y = coordinates

    rest_x = abs_x % SCREEN_WIDTH
    rest_y = abs_y % SCREEN_HEIGHT

    unscrolled_coordinates = (rest_x, rest_y)

    return unscrolled_coordinates


def get_straight_distance(point_a, point_b):
    """Calculate the straight distance between two points.

    get_straight_distance(point_a, point_b) -> float

    point_a, point_b
        Each is represented by a list or tuple with two
        values: x and y respectively. For instance: (x, y).

    Illustration:
                 ._ _
        (point a)|\   |
                 | \s |
              dy |  \ |
                 |   \|
                 |_ _ .(point b)
                   dx

        where:
        s  = the straight distance  (straight_distance)
        dy = the distance on y axis (distance_x)
        dx = the distance on x axis (distance_y)

    >>> a, b = (10, 10), (20, 20)
    >>> round(get_straight_distance(a, b), 2)
    14.14
    """
    x_a, y_a = point_a
    x_b, y_b = point_b

    distance_x = x_b - x_a
    distance_y = y_b - y_a

    straight_distance = hypot(distance_x, distance_y)

    return straight_distance


# XXX
# Both offset_point and invert_point functions
# could be further abstracted in a Point or Vector2d class.
# I've already faced some situations where such
# associations were quite intuitive, so I just maybe
# I could see into it in the near future.
# If so, and only if, probably the existing
# pygame.math.Vector2d class
def offset_point(point, offset):
    """Return offset point (an int tuple of length 2).

    offset_point(point, offset) -> point

    point
        A list or tuple containing two integers representing
        a point's coordinates in space.
    offset
        A list or tuple containing two integers representing
        amounts to be added to each point coordinate.

    >>> offset_point((10, 10), (5, 5))
    (15, 15)
    >>> offset_point((10, 10), (-5, 5))
    (5, 15)
    >>> offset_point((10, 10), (5, -5))
    (15, 5)
    >>> offset_point((10, 10), (-5, -5))
    (5, 5)
    """
    x, y = point
    x_offset, y_offset = offset

    offset_point = [x + x_offset, y + y_offset]

    return offset_point


def invert_point(point, invert_x, invert_y):
    """Return point with inverted coordinates as requested.

    Inverting a coordinates means changing the signal.

    point
        A list or tuple with two integers representing
        x and y coordinates of a point.
    invert_x
    invert_y
        Booleans indicating whether the x and y coordinates
        should be inverted, respectively.

    >>> invert_point((10, 10), True, True)
    (-10, -10)
    >>> invert_point((10, 10), True, False)
    (-10, 10)
    >>> invert_point((10, 10), False, True)
    (10, -10)
    >>> invert_point((10, 10), False, False)
    (10, 10)
    """
    x, y = point
    new_x = -x if invert_x else x
    new_y = -y if invert_y else y
    inverted_point = (new_x, new_y)

    return inverted_point



"""Facility for game resource gathering and distribution."""

from os.path import join

### local imports

from ..config import ANIMATIONS_DIR

from ..common.jsonhandler import load_json

from .anim.proc.main import process_anim_data


### gather animation resources

ANIM_DATA_MAP = {}


anim_files_map = {
    "dani": "dani.ani2d",
    "cloud": "cloud.ani2d",
    "mushroom": "mushroom.ani2d",
    "indie_python_logo": "indie_python_logo.ani2d",
    "thorn_shooter": "thorn_shooter.ani2d",
    "satellite_dish": "satellite_dish.ani2d",
    "rocket_scorpion": "rocket_scorpion.ani2d",
    "toxic_pollen_cloud": "toxic_pollen_cloud.ani2d",
    "grass_and_dirt_block": "grass_and_dirt_block.ani2d",
}


for key, filename in anim_files_map.items():

    json_data = load_json(join(ANIMATIONS_DIR, filename))

    try:
        ANIM_DATA_MAP[key] = process_anim_data(json_data)

    except Exception as err:

        print("Error while processing {} file.".format(filename))
        raise Exception from err



"""Facility for color definitions and related tools.

Variable names for colors either describe the color names
(for instance, BLACK, WHITE, etc.), approximate color names
(for instance, SHALIMAR, which doesn't describe the SHALIMAR
color, but it is very close and don't have a specific name, so
we use SHALIMAR instead) or their roles in the package
(for instance, PRIMARY_COLOR_MEDIUM).
"""


### Neutral colors
BLACK = (0, 0, 0)
WHITE = HIGHLIGHT = (255, 255, 255)
SHADOW = (102, 102, 102)

GRAY_LIGHTER = (212, 212, 212)
GRAY_LIGHT = (170, 170, 170)
GRAY_MEDIUM = (128, 128, 128)
GRAY_DARK = (84, 84, 84)
GRAY_DARKER = (42, 42, 42)

ARBITRARY_COLORKEY = (129, 129, 129)

### Game art pallete and aliases
# (aliases are color names for approximate colors)
PRIMARY_COLOR_LIGHTER = SHALIMAR = (255, 253, 195)
PRIMARY_COLOR_LIGHT = WITCH_HAZE = (255, 251, 155)
PRIMARY_COLOR_MEDIUM = CHENIM = (227, 222, 100)
PRIMARY_COLOR_DARK = EARLS_GREEN = (193, 188, 63)
PRIMARY_COLOR_DARKER = LEMON_GINGER = (159, 154, 31)

SECONDARY_COLOR01_LIGHTER = PASTEL_GREEN = (114, 222, 121)
SECONDARY_COLOR01_LIGHT = EMERALD = (72, 206, 81)
SECONDARY_COLOR01_MEDIUM = FOREST_GREEN = (36, 193, 47)
SECONDARY_COLOR01_DARK = SALEM = (10, 170, 20)
SECONDARY_COLOR01_DARKER = JAPANESE_LAUREL = (1, 133, 9)

SECONDARY_COLOR02_LIGHTER = DEEP_BLUSH = (218, 92, 161)
SECONDARY_COLOR02_LIGHT = MEDIUM_RED_VIOLET = (207, 54, 137)
SECONDARY_COLOR02_MEDIUM = LIPSTICK = (201, 6, 113)
SECONDARY_COLOR02_DARK = FRESH_EGGPLANT = (159, 1, 87)
SECONDARY_COLOR02_DARKER = SIREN = (125, 0, 68)

COMPLEMENT_COLOR_LIGHTER = PICTON_BLUE = (44, 184, 237)
COMPLEMENT_COLOR_LIGHT = CERULEAN = (7, 178, 244)
COMPLEMENT_COLOR_MEDIUM = CERULEAN_02 = (1, 153, 211)
COMPLEMENT_COLOR_DARK = ALLPORTS = (1, 111, 154)
COMPLEMENT_COLOR_DARKER = ORIENT = (1, 87, 119)

### Other colors for specific usage
SKY_COLOR = (108, 166, 205)
DIALOGUEBOX_BACKGROUND = (255, 246, 143)
PROMPT_BACKGROUND = (84, 255, 159)
SIGN_POLE_COLOR = (185, 185, 185)
ORANGE = (255, 128, 0)
RED = (255, 0, 0)


def has_same_rgb(color1, color2):
    """Return True if rgb values are the same.

    The advantage of this function is that it doesn't
    care if color are represented by lists or tuples,
    as long as their respective rgb values have the same
    index.

    Without this functions, using
    [x, y, z] == (x, y, z) would equal False, since
    though their values are the same, they have different
    types. This would be an undesired answer, since those
    values represent the same color and thus are equal.

    Also, this functions works with color objects from
    pygame.color.Color.

    >>> has_same_rgb([22, 33, 44], [22, 33, 44])
    True
    >>> has_same_rgb([22, 33, 44], [ 2, 33, 44])
    False
    >>> has_same_rgb([22, 33, 44], (22, 33, 44))
    True
    >>> has_same_rgb([22, 33, 44, 12], [22, 33, 44])
    True
    """
    return all(color1[i] == color2[i] for i in range(3))



"""Facility for simple object display."""

### local imports

from ..config import SCREEN_RECT

from ..common.behaviour import empty_function

from ..appcommon.surf import INVISIBLE_SURF, render_image

from ..appcommon.anim.player.main import AnimationPlayer

from .proc import process_tasks


class AnimatedObject:
    """Animated object to hold animation."""

    def __init__(
        self,
        anim_data_key,
        tasks,
        exchange_pos_attrs=("center", "center"),
        offset=(0, 0),
    ):
        """Create attributes."""
        ### retrieve coordinates name and value for
        ### animated object

        rect_attr, screen_rect_attr = exchange_pos_attrs

        pos_value = getattr(SCREEN_RECT.move(offset), screen_rect_attr)

        ### instantiate animation player

        self.anim_player = AnimationPlayer(
            self, anim_data_key, coordinates_value=pos_value, coordinates_name=rect_attr
        )

        self.anim_player.play()

        ### process tasks
        process_tasks(self, tasks)

    def update(self):
        """Update animation player."""
        self.anim_player.update()

    def draw(self):
        """Call animation player draw method."""
        self.anim_player.draw()



"""FPS display facility to assist in game development."""

### local imports

from .config import SCREEN_RECT, CLOCK
from .palette import BLACK, WHITE

from .appcommon.text.label.autolabel import AutoLabel

### define function to get current frames per second
### custom formatted


def custom_formatted_fps():
    """Retrieve, format and return frames/second rate."""
    rounded_fps = round(CLOCK.get_fps())

    custom_formatted_fps = str(rounded_fps).rjust(2, "0")

    return custom_formatted_fps


### create autolabel

fps_display = AutoLabel(
    custom_formatted_fps,
    "FPS: ",
    "",
    text="30",
    font_size=18,
    foreground_color=WHITE,
    background_color=BLACK,
    coordinates_name="topright",
    coordinates_value=SCREEN_RECT.topright,
)



from os import listdir, mkdir
from os.path import join
from shutil import copyfile

### local imports

from ...logconfig import APP_LOGGER

from ...config import (
    GAME_REFS,
    SAVES_DIR,
    STATE_FILE_NAME,
    DEFAULT_STATE_DATA,
    IMAGES_DIR,
)

from ...common.timehandler import get_isostring_datetime
from ...common.jsonhandler import save_json


### instantiate local logger
logger = APP_LOGGER.getChild(__name__)


def save_game():
    """Save game by copying buffer to original files.

    All changes made in the game and levels are
    store in buffers whenever the player switch levels,
    progresses, etc. Saving the game just means copying
    those buffers to the original files, so they can
    be loaded with the progres/changes applied.
    """
    logger.info("Saving game")

    save_dir = GAME_REFS.dirname

    ### Saving state

    state_path = join(SAVES_DIR, save_dir, STATE_FILE_NAME)

    state_swap_path = state_path + ".swp"

    try:
        copyfile(state_swap_path, state_path)

    except FileNotFoundError as err:
        msg = "Didn't found state swap file" + " in " + save_dir + " save."
        logger.exception(msg)
        raise err

    ### Saving levels

    levels_dir = join(SAVES_DIR, save_dir)

    level_files = [
        join(SAVES_DIR, save_dir, item)
        for item in listdir(levels_dir)
        if item.endswith(".lvl")
    ]

    for level_path in level_files:

        level_swap_path = "{}.swp".format(level_path)

        try:
            copyfile(level_swap_path, level_path)

        except FileNotFoundError as err:

            # don't raise this error, it must only be logged,
            # since it is natural; not all levels
            # will have a swap version (when the player
            # saves elsewhere and don't visit there again),
            # that's why some of them cause the
            # file not found error to be raised.

            # XXX
            # This error can happen when there's no
            # swap file because I saved it (so swaps
            # were eliminated). However, I'm somehow
            # dissatisfied with how I'm logging it,
            # it isn't clear. Solve this.

            msg = (
                "Didn't found level swap path"
                + " called "
                + level_swap_path
                + " in "
                + save_dir
                + " save."
            )

            logger.exception(msg)


def setup_new_save(directory_name):
    """Setup all files for a new game save/slot."""
    logger.info("Setting up a new save")

    new_save_dir = join(SAVES_DIR, directory_name)
    mkdir(new_save_dir)

    state_data = DEFAULT_STATE_DATA.copy()
    state_data["last_played_datetime"] = get_isostring_datetime()

    state_filepath = join(new_save_dir, STATE_FILE_NAME)
    save_json(state_data, state_filepath)

    thumb_file = join(IMAGES_DIR, "new_save_thumb.png")
    thumb_destination = join(new_save_dir, "thumb.png")
    copyfile(thumb_file, thumb_destination)



"""Facility for grass and dirt block sprite."""

from functools import partial
from random import seed, randint
from types import SimpleNamespace

from ..appcommon.anim.player.main import AnimationPlayer
from ..appcommon.task import add_task
from ..common.math import get_straight_distance


seed()

### utility function

# pygame don't handle lambdas well, so using this
def get_distance(pos, item):
    """Return straight distance relative to pos."""
    return get_straight_distance(item.rect.midtop, pos)


### class definition


class GrassDirtBlock:
    """A block with dirt and grass.

    The grass responds to player movement when walking
    over it."""

    level = None

    def __init__(self, prop_name, coordinates):
        """Initialize superclass and variables."""
        self.prop_name = prop_name

        self.anim_player = AnimationPlayer(
            self,
            prop_name,
            coordinates_name="bottomleft",
            coordinates_value=coordinates,
        )

        self.perform_setups()

    def perform_setups(self):
        """Perform extra setups.

        Those setups are either extra processing done
        to embellish in game stuff or controls, like
        flags to keep track of states.
        """
        self.patch_task = SimpleNamespace(finished=True)

        # XXX
        # in the future, this could have a larger range,
        # but I'd need more surfaces. The grass would be
        # more varied and beautiful.
        steps_no = randint(0, 1)

        obj_name_list = [
            obj.name for obj in self.anim_player.get_nodes(exclude_root=True)
        ]

        self.anim_player.update_nodes(steps_no, obj_name_list)

        ### store the nodes representing the grass patches

        self.patches = [
            obj
            for obj in self.anim_player.get_nodes()
            if obj.name.startswith("grass_patch")
        ]

    def update(self):
        """Update animation."""
        self.anim_player.update()

    def draw(self):
        """Draw animation."""
        self.anim_player.draw()

    def select_stomped_grass(self, pos):
        """Return name of grass patch closer to position.

        This is used to change the surface of the grass
        to make it appear as if it were stomped and thus
        changed shape slightly.

        pos
            midbottom position of player rect.
        """
        func = partial(get_distance, pos)
        self.patches.sort(key=func)
        return self.patches[0].name

    def grass_walking(self, player, pos):
        """Trigger grass movement if player walks over."""
        if player.x_speed and self.patch_task.finished:
            self.set_grass_moving(pos)

    def set_grass_moving(self, pos):
        """Set grass moving near position.

        pos
            midbottom position of player rect.
        """
        obj_name = self.select_stomped_grass(pos)

        callable_task = partial(self.anim_player.update_nodes, 1, [obj_name])

        self.patch_task = add_task(callable_task, 25)

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)



"""Facility for processing animation data."""

### local imports

## "common" subpackage

from ....common.wdeque.main import WalkingDeque
from ....common.pointgen import generate_points, convert_items_to_tuples
from ....common.indexgen import generate_indices
from ....common.tree import get_tree_values, get_tree_height

## "appcommon" subpackage

from ...surf import INVISIBLE_SURF
from ...gensurf import generate_surfaces

# animation data processing assistance

from .assist import (
    get_filled_map,
    manage_proxy_animations,
    store_anim_clock_keys,
    store_root_pos_exchange_map,
)


def process_anim_data(data):
    """Return processed animation data.

    data (dict)
        Data loaded from an .ani2d JSON file.

    The animation data is processed in several steps and
    returned as a mapping of processed data for different
    purposes.

    An explanation of the animation data model, as well as
    input, processing and output data of this preprocessing
    facility can be found in the anim_data_model.rst file
    here:

    appcommon/anim/anim_data_model.rst

    ### discussion about animation values conversion into
    ### tuples in other packages

    Different from what happens in this function
    here in the animation editor, the same function
    when implemented in the game package (should) converts
    the final animation values (including the value maps)
    into tuples. This is so because, as profiled by me and
    explained and confirmed by a prominent professional
    in the Python community (Raymond Hettinger), here:

    http://stackoverflow.com/a/22140115/10194508

    tuples have a slight boost in speed for item access (as
    well as other less important but welcome optimizations
    in other areas). It might not be a very significant
    change, but it is implemented anyway because:

    1) it wasn't costly in the point of view of management/
    maintenance (it's just one additional line of code);
    2) even not being very significant, it is expected that
    the combined speed gains of multiple solutions over the
    time, including this one, may prove to make a difference;

    However, such implementation wasn't made here in the
    animation editor. This is so because we have lots of
    read/edit/write operations, and tuples would need a lot
    of time consuming conversions to list and back into
    tuples again. Thus here in the animation editor the final
    animation value lists are never converted to tuples when
    they are processed.
    """
    ### process and post-process animation value maps

    ## create surface map and fill it with commonly used
    ## surface sequences (if the raw data in the maps
    ## is provided)

    surf_map = get_filled_map(
        {}, data, "surface_map", "surface_map_post", generate_surfaces
    )

    # also convert surface map values (lists) into tuples
    surf_map.update({key: tuple(value) for key, value in surf_map.items()})

    ## create position map and fill it with commonly used
    ## position sequences (if the raw data in the maps
    ## is provided)

    pos_map = get_filled_map(
        {}, data, "position_map", "position_map_post", generate_points
    )

    # also guarantee all points in positions lists are tuples
    # (this is also perform when processing animation values
    # further ahead)
    for position_list in pos_map.values():
        convert_items_to_tuples(position_list)

    # also convert position map values (lists) into tuples
    pos_map.update({key: tuple(value) for key, value in pos_map.items()})

    ### process animation data

    ## retrieve animation values for processing
    raw_anim_values = data["values"]

    ## create map wherein to store the results of the
    ## processed animation values (this will be be surfaces
    ## and positions mapped to node names and further
    ## mapped to animation names
    values = {}

    ## list names of the two types of values processed
    value_types_names = ["surfaces", "positions"]

    ## create map associating each type of value with its
    ## related variable map

    variables_map = {
        "surfaces": {
            "raw_data_key": "surface_data",
            "proc_data_key": "surfaces",
            "default": INVISIBLE_SURF,
            "value_map": surf_map,
            "function": generate_surfaces,
        },
        "positions": {
            "raw_data_key": "position_data",
            "proc_data_key": "positions",
            "default": (0, 0),
            "value_map": pos_map,
            "function": generate_points,
        },
    }

    ## iterate over animation values for each animation
    for anim_name, anim_data in raw_anim_values.items():

        ## create a dict to hold values for this specific
        ## animation
        values[anim_name] = {}

        ## iterate further over animation values for each
        ## node in this animation

        for node_name, node_data in anim_data.items():

            ## create a dict to hold values for this node
            ## in this animation

            node_anim_values = values[anim_name][node_name] = {}

            # use each variable map to retrieve important
            # variables and use them to process data

            for value_type in value_types_names:

                var_map = variables_map[value_type]

                raw_data_key = var_map["raw_data_key"]
                proc_data_key = var_map["proc_data_key"]
                default = var_map["default"]
                value_map = var_map["value_map"]
                func = var_map["function"]

                # try retrieving data
                try:
                    value_data = node_data[raw_data_key]

                # if absent, use default

                except KeyError:

                    node_anim_values[proc_data_key] = [default]

                    # if we are dealing with positions
                    # use the raw_data_key to assign
                    # the default value to the node
                    # data mapping as well (which is the
                    # mapping to be saved in the file)
                    if value_type == "positions":
                        node_data[raw_data_key] = [default]

                # otherwise process and store values;
                # note: in game applications we convert
                # the values container to a tuple before
                # storing them; see the docstring to know
                # more;

                else:

                    # process data into values
                    processed_values = func(value_data, value_map)

                    # if we are dealing with positions, we
                    # should convert the positions within
                    # the list into tuples

                    if raw_data_key == "position_data":

                        try:
                            convert_items_to_tuples(processed_values)

                        except TypeError:
                            pass

                    # regardless of the type of values
                    # being handled, they must be
                    # converted into tuples for quicker
                    # indexing
                    processed_values = tuple(processed_values)

                    # store values
                    node_anim_values[proc_data_key] = processed_values

    ### pre-process timing data

    ## create timing map and fill it with commonly used
    ## index sequences (if the raw data in the maps
    ## is provided)

    timing_map = get_filled_map(
        {}, data, "timing_map", "timing_map_post", generate_indices
    )

    ### process timing data

    ## retrieve timing data for processing
    raw_timing_data = data["timing"]

    ## create map wherein to store the results of the
    ## processed timing data (this will be
    ## collections.deque's subclass instances used
    ## to hold indices for surfaces and positions
    ## mapped to node names and further mapped to
    ## animation names
    timing = {}

    ## iterate over timing data for each animation
    for anim_name, anim_data in raw_timing_data.items():

        ## create dict to hold timing data for this specific
        ## animation
        timing[anim_name] = {}

        ## iterate further over timing data for each
        ## node in animation
        for node_name, node_timing_data in anim_data.items():

            ## create dict to hold timing data for this
            ## specific node in this animation
            node_timing_values = timing[anim_name][node_name] = {}

            # process data to retrieve timing values
            # for surfaces and positions
            for key in ("surface_indices", "position_indices"):

                # try retrieving data
                try:
                    timing_data = node_timing_data[key]

                # if absent, use default
                except KeyError:
                    node_timing_values[key] = WalkingDeque([0])

                    # also use the key to assign the
                    # default value to the node timing
                    # data in the file
                    node_timing_data[key] = [0]

                # otherwise, process and store values
                else:

                    # get indices
                    indices = generate_indices(timing_data, timing_map)

                    # create a walking deque with them
                    node_timing_values[key] = WalkingDeque(indices)

    ### process structure data

    structure = data["structure"]

    ## pre-assign structures from map

    # try retrieving structure mapped data
    try:
        struct_map = data["structure_map"]

    # if absent, pass
    except KeyError:
        pass

    # otherwise, retrieve data from such map and assign
    # wherever requested
    else:

        for anim_name, anim_struct_data in structure.items():

            if isinstance(anim_struct_data, str):

                key = anim_struct_data

                struct_data = struct_map[key]

                structure[anim_name] = struct_data

    ## iterate over structures processing and storing:
    ## - default ordering for updating and drawing as needed
    ## - list of nodes in the animation
    ## - the height of the animation tree

    for anim_name, anim_struct_data in structure.items():

        # retrieve tree for this animation
        tree = anim_struct_data["tree"]

        # generate and store optional order (parents first)
        optional_order = get_tree_values(tree, "name", "children")

        # assign the optional order for key, only if missing

        for order in ("updating_order", "drawing_order"):

            try:
                anim_struct_data[order]

            except KeyError:
                anim_struct_data[order] = optional_order

        # also store the optional order in a special
        # "node_names" field, because regardless of whether
        # it is or not in the same order of the other
        # ordering fields, its role is to list all nodes
        # in the animation, regardless of order
        anim_struct_data["node_names"] = optional_order

        # store height of this animation's tree
        anim_struct_data["height"] = get_tree_height(tree)

    ### process animation metadata

    ## retrieve animation metadata for processing, if any
    metadata = data.get("metadata", {})

    ## define which animations are of the proxy type,
    ## storing their names in the metadata dict;
    ## we do so by build a list of keys from the structure
    ## dict which are not present in the animation values
    ## or timing dict; such keys represent animation names
    ## of proxy type

    values_and_timing_keys = (*values, *timing)

    metadata["proxy_anim_names"] = [
        anim_name for anim_name in structure if anim_name not in values_and_timing_keys
    ]

    ## manage proxy animations and related metadata
    manage_proxy_animations(
        metadata, structure, values, timing, surf_map, pos_map, timing_map
    )

    ## generate metadata about durations
    store_anim_clock_keys(metadata, timing)

    ## generate/gather metadata about root position exchange
    store_root_pos_exchange_map(metadata, structure)

    ### process nodes data

    nodes = data["nodes"]

    # try retrieving geometry mapped data
    try:
        geometry_map = data["geometry_map"]

    # if absent, pass
    except KeyError:
        pass

    # otherwise, retrieve data from such map and assign
    # wherever requested
    else:

        for node_data in nodes.values():

            geometry_data = node_data["geometry"]

            if isinstance(geometry_data, str):
                key = geometry_data
                node_data["geometry"] = geometry_map[key]

    ### Map collections and return resulting map

    anim_data = {
        "values": values,
        "timing": timing,
        "structure": structure,
        "metadata": metadata,
        "nodes": nodes,
    }

    return anim_data



"""Facility for the thorn shooter bullet."""

from math import hypot

from ...appcommon.autoblit import BasicObject
from ...appcommon.surf import render_rect
from ...appcommon.collision import does_touch_screen
from ...common.math import unscroll_coordinates


class ThornBullet(BasicObject):
    """A moderately slow thorn bullet."""

    def __init__(self, center, level):
        """Initialize superclass and variables."""
        self.level = level
        self.image = render_rect(5, 5, (255, 0, 0))
        self.rect = self.image.get_rect()

        unscrolled_center = unscroll_coordinates(center)
        self.rect.center = self.center = unscrolled_center

        self.speed = 7  # px/frame
        self.power = 10

        self.is_targeting = False

    def update(self):
        """Update object state."""
        if not self.is_targeting:
            self.acquire_target()
            self.is_targeting = True

        self.move()

    def acquire_target(self):
        """Get player direction.

        Obs.: though part of the calculations could be
        performed by get_straight_distance function in
        common.behaviour, we need access to distance
        x and y, which said function don't provide. So
        this is good as it is."""
        self.target_x, self.target_y = self.level.player.rect.center
        x_distance = self.target_x - self.rect.x
        y_distance = self.target_y - self.rect.y

        diagonal_distance = hypot(x_distance, y_distance)

        time_in_frames = diagonal_distance / self.speed

        self.x_speed = round(x_distance / time_in_frames)
        self.y_speed = round(y_distance / time_in_frames)

    def move(self):
        """Perform movement in x and y direction."""
        self.rect.x += self.x_speed
        self.rect.y += self.y_speed

        collided_platform = self.level.onscreen_map["blocks"].collide_any(self)

        if collided_platform:
            self.reset_position()
            return

        is_on_screen = does_touch_screen(self)

        if not is_on_screen:
            self.reset_position()
            return

        hit_player = self.rect.colliderect(self.level.player.rect)

        if hit_player:

            self.reset_position()

            try:
                self.level.player.suffer_damage(self.power)

            except AttributeError:
                pass

            return

    def reset_position(self):
        """Reset bullet position."""
        self.rect.center = self.center
        self.is_targeting = False



"""Facility for application specific behaviours."""

from os.path import join
from shutil import copyfile

### local imports

from ...logconfig import APP_LOGGER

from ...config import GAME_REFS, LEVELS_DIR, SAVES_DIR, STATE_FILE_NAME

from ...common.jsonhandler import load_json, save_json

from .load import load_game
from .buffer import remove_buffers

from ...mainmenu.main import MainMenu
from ...gamelevel.main import GameLevel
from ...sceneplayer.main import ScenePlayer


### instantiate local logger
logger = APP_LOGGER.getChild(__name__)


### functions


def get_level(level_name):
    """Return a new GameLevel instance.

    level_name (string)
        represents the basename of the level file. As level
        files are designed now, that string will end with
        the '.lvl' extension.
    """
    save_dir = GAME_REFS.dirname

    # destination, when present, holds a string with the
    # name of a portal in the level whose coordinates
    # are used to position the player. Whenever
    # entering a new level via a portal a specific
    # portal destination will be provided.

    try:
        destination = GAME_REFS.destination
    except AttributeError:
        destination = None
    else:
        del GAME_REFS.destination

    msg = "Level named '{}' being loaded.".format(level_name)

    logger.debug(msg)

    level_filepath = join(SAVES_DIR, save_dir, level_name)

    state_filepath = join(SAVES_DIR, save_dir, STATE_FILE_NAME)

    try:
        new_level = GameLevel(level_filepath, state_filepath, destination)

    except FileNotFoundError:

        backup_level_filepath = join(LEVELS_DIR, level_name)

        copyfile(backup_level_filepath, level_filepath)

        new_level = GameLevel(level_filepath, state_filepath, destination)

    return new_level


def get_custom_manager(manager_name):
    """Return custom update manager instance."""
    if manager_name.endswith(".scn"):
        return ScenePlayer(manager_name)

    elif manager_name == "main_menu":
        return MainMenu()

    else:
        raise Exception("This block should never be executed.")


def reset_player_health():
    """Reset player health from save_dir slot.

    Designed to be used after player dies,
    before reloading the last save/state in current
    slot.
    """
    logger.info("Health reset.")

    save_dir = GAME_REFS.dirname

    state_path = join(SAVES_DIR, save_dir, STATE_FILE_NAME)

    state_data = load_json(state_path)
    state_data["health"] = 100
    save_json(state_data, state_path)



"""Facility for animation playback."""

from copy import deepcopy
from functools import partialmethod

### third-party import
from pygame.draw import rect as draw_rect

### local imports

from ....config import SCREEN

from ..node import Node
from ...resources import ANIM_DATA_MAP

from ....palette import ORANGE, RED

## animation player class extension
from .utils import Utilities


# XXX maybe: turn the anim_name attribute into a property,
# making it so that setting it causes the set animation
# behaviours to be executed (after verifying if animation
# name exists); ponder; I'm not fond of this though,
# because this kind of setup means extra overhead for
# real time processing (even if just a little);


class AnimationPlayer(Utilities):
    """Plays animation.

    This object receives a filepath and passes it to
    a factory function used to generate animation
    data. The animation data is then used to instantiate
    support objects called nodes which are arranged in
    tree-like structures to play different animations
    which consist of objects having surfaces and/or
    positions constantly changed to convey the illusion
    of animation.

    The animation manager is part of the .ani2d interface,
    which consists of a data model and a set of modules
    and applications to allow easy and modular animation
    management (creation, edition and playback). The .ani2d
    interface is quite extensive, making it unpractical to
    cover everything here.

    Proper comprehensive documentation is provided here:
    appcommon/anim/ani2d_interface.rst
    """

    def __init__(self, game_obj, anim_data_key, coordinates_value, coordinates_name):
        """Store data and perform setups.

        game_obj (any 'object' subclass)
            object to hold the root node rect, that is, the
            'link' between the game/application world and the
            animation manager.
        anim_data_key (string)
            represents a key to retrieve the .ani2d
            pre-processed animation data from a map.
        coordinates_value (2-tuple or list with 2 integers)
            holds the x and y value of a position in
            two dimensions.
        coordinates_name (string)
            attribute name of a pygame.Rect instance wherein
            to store the coordinates_value (for instance
            "topleft").
        """
        ### store game_obj reference
        self.game_obj = game_obj

        ### setup animation data and nodes

        ## retrieve the pre-processed animation data
        ## using the provided key
        anim_data = ANIM_DATA_MAP[anim_data_key]

        ## retrieve and store animation data

        self.values = anim_data["values"]
        self.structure = anim_data["structure"]
        self.metadata = anim_data["metadata"]
        self.nodes_data = anim_data["nodes"]

        # the timing data must be deepcopied, though;
        # the animation timing is what makes each instance
        # independent from the others;
        self.timing = deepcopy(anim_data["timing"])

        ## create a mapping with animation nodes

        self.node_map = {
            node_name: Node(node_name, node_data)
            for node_name, node_data in self.nodes_data.items()
        }

        ## create a set of animation names
        self.anim_names = set(self.structure.keys())

        ### set default behaviours

        self.pause()
        self.set_draw_routine("draw_nodes")

        ### set current animation

        ## choose animation name and switch to it

        # pick either "idle", or the first one from a
        # tuple with the animation names
        anim_name = "idle" if "idle" in self.anim_names else (*self.anim_names,)[0]

        # then switch to the chosen animation
        self.switch_animation(anim_name)

        ### position game_obj

        setattr(self.game_obj.rect, coordinates_name, coordinates_value)

    ### animation switching support

    def switch_animation(self, animation_name):
        """Switch current animation to requested one.

        Also return True if succeed.

        animation_name (string)
            represents the name of an animation.

        If the animation name provided doesn't exist,
        just print the error and do nothing.
        """
        ### check if the animation structure data is present
        ### (in practice, there are other requirements for
        ### an animation to be deemed playable, but since
        ### those are already pre-checked in the animation
        ### data processing, we do a very simplified checkup
        ### here for the sake of efficiency)
        try:
            self.structure[animation_name]

        ### if absent, then deem the data to play the
        ### animation insufficient
        except KeyError:
            print(
                "Didn't manage to find enough data to play",
                "the animation called",
                animation_name,
            )

        ### otherwise, perform the needed setups to leave
        ### the animation ready for being played and
        ### return True
        else:

            ## store the animation name
            self.anim_name = animation_name

            ## clear the rotation in all timing indices,
            ## to guarantee we are starting from the first
            ## index of animation
            self.clear_rotations()

            ## make all needed setups to ensure the
            ## animation follows it structure
            self.set_root_and_structure()

            ## update nodes without walking the animation
            ## values (that is, without passing a step
            ## argument), to ensure the animation values
            ## for the first frame are assigned to the
            ## nodes
            self.update_nodes()

            ## store an animation clock to use as reference
            ## for the animation timing
            self.store_animation_clock()

            ## admin task: check need to skip walking
            ## out of the first frame when updating
            self.check_walking_skipping()

            ## return True to indicate success
            return True

    def clear_rotations(self):
        """Restore timing indices to first frame.

        This is done by calling a 'restore_walking' method
        on a collections.deque subclass.
        """
        ### retrieve timing data for current animation
        timing_data = self.timing[self.anim_name]

        ### iterate over each node clearing the rotation
        ### for both surface and position time indices

        for node_name in timing_data:
            node_timing = timing_data[node_name]

            for key in ("surface_indices", "position_indices"):
                node_timing[key].restore_walking()

    def set_root_and_structure(self):
        """Setup root node and animation structure.

        In this method we perform three important operations
        related to the animation structure:

        (1) we check the need to reposition the new root
            in the application/game world and the need to
            set a new rect for the game object;
        (2) we set references in each node's "parent"
            attribute to their respective parent node.
            The root node, which doesn't have a parent,
            has None assigned to the parent attribute;
        (3) we also store structure related data into
            special attributes for usage when referencing
            the animation nodes used on the current
            animation;

        Below you find more details about operation (1) and
        (2). There was no need to further comment the
        operation (3).

        About operation (1):

        The game object references one of the root rects
        which is how the object moves in the application
        or game world. We switching between different
        animations, the root might be a different object,
        in which case position information is exchanged
        between them to guarantee the object stays in the
        same spot and the game object references the
        new root's rect.

        About operation (2):

        Just by setting references to their parent nodes
        in each child node we guarantee that the tree
        structure takes effect, because each node
        automatically repositions itself relative to its
        parent when being updated on the update_nodes method.
        """
        ### reference structure data for the current
        ### animation and its tree data

        anim_structure = self.structure[self.anim_name]
        tree = anim_structure["tree"]

        ### check the need to reposition the new root
        ### in the application/game world and the need to
        ### set a new rect for the game object

        ## retrieve new root
        new_root = self.node_map[tree["name"]]

        ## also reference its relevant rect (either the
        ## collision rect, if present, or the normal rect);
        ## we call it root rect, because all nodes are
        ## positioned relative to it;
        root_rect = new_root.col_rect or new_root.rect

        ## try retrieving previous root (if any)
        try:
            previous_root = self.root

        ## if there weren't one, just assign the root rect
        ## to the game obj rect attribute
        except AttributeError:
            self.game_obj.rect = root_rect

        ## if a previous root exists, verify the need to
        ## update the position of the new root with the
        ## position information of the previous root
        else:

            if new_root is not previous_root:
                self.exchange_root_position(previous_root, new_root)

                # also assign the root rect to the game obj
                # rect attribute
                self.game_obj.rect = root_rect

        ## store the new root node in the 'root' attribute,
        ## which represents the current root being used
        self.root = new_root

        ## also store the root rect in a special attribute;
        ## for convenience
        ## (note: this reference itself isn't used in this
        ## class, but it is useful for things like measuring
        ## distances between the root_rect and nodes;
        ## such measurements are performed for instance, by
        ## an external tool used to export the current
        ## animation to a sequence of .png files)
        self.root_rect = root_rect

        ### set references in each node's "parent"
        ### attribute to their respective parent node

        self.root.parent = None
        self.set_parent_references(tree)

        ### store updating order, drawing order and
        ### node_names list

        for key in ("updating_order", "drawing_order", "node_names"):
            setattr(self, key, anim_structure[key])

    def exchange_root_position(self, previous_root, new_root):
        """Exchange position from previous to new root.

        previous_root, new_root
        (instances of appcommon.anim.node.Node)
            previous root from which to retrieve the
            position information used and new root to have
            its position updated.
        """
        ### retrieve exchange map
        exchange_map = self.metadata["root_pos_exchange_map"]

        ### retrieve pygame.Rect attribute names used to
        ### exchange the position data

        attribute_names = exchange_map[previous_root.name][new_root.name]

        prev_attr_name, new_attr_name = attribute_names

        ### retrieve rects from previous and new root with
        ### preference for the collision rect

        new_rect = new_root.col_rect or new_root.rect

        prev_rect = previous_root.col_rect or previous_root.rect

        ### finally, retrieve the position value from
        ### previous root rect and apply to new root rect

        pos = getattr(prev_rect, prev_attr_name)
        setattr(new_rect, new_attr_name, pos)

    def set_parent_references(self, tree):
        """Set parent references on nodes.

        A tree structered mapping data is used to retrieve
        parent nodes and their children and while doing
        so, create references for each parent.

        tree (dict)
            Contains node data. The data may have a
            'children' field listing more nodes' data,
            in which case it is processed recursively.
        """
        ### retrieve node reference using the value in the
        ### tree's 'name' key
        node = self.node_map[tree["name"]]

        ### check if it has children
        try:
            children = tree["children"]

        ### if not just pass
        except KeyError:
            pass

        ### otherwise, iterate over each child data
        ### mapping stored in the children list,
        ### setting parent attributes on each
        ### child pointing to the current node;
        ### also check each child recursively
        ### for further parent relationships down
        ### the tree;
        else:

            for child_data in children:

                ### set parent attribute on child

                child = self.node_map[child_data["name"]]
                child.parent = node

                ### check child data recursively for
                ### more parent-children relationships
                self.set_parent_references(child_data)

    def store_animation_clock(self):
        """Store a reference of an animation clock.

        An animation clock is just an instance of
        the common.walkingdeque.WalkingDeque class,
        a collections.deque (from python standard library)
        subclass. The name here implies its role, since we
        use it to control animation timing.

        For more info, head to the module itself:
        common/walkingdeque.py
        """
        ### retrieve a mapping containing data about the
        ### largest deque for the current animation
        data = self.metadata["largest_deques"][self.anim_name]

        ### from the data, retrieve the name of the node
        ### and sequence from which to retrieve the
        ### animation clock
        node_name = data["node_name"]
        sequence_name = data["sequence_name"]

        ### finally retrieve and store the animation clock
        self.anim_clock = self.timing[self.anim_name][node_name][sequence_name]

    ### update routines

    def update_nodes(self, steps=0, nodes=None):
        """Rotate and assign animation values to nodes.

        The rotation per se is performed in the timing
        indices which point to the animation values,
        which are them assigned/passed to the nodes via
        the image attribute for surfaces and the set_pos
        method for positions.

        steps (int, defaults to 0)
            How much to rotate the data. steps=0
            means no rotation. rotate < 0 means rotate forth
            and rotate > 0 means rotate back.
        nodes (iterable containing strings)
            Names for nodes you want to update.

        The default value of "steps" makes the same
        values (surfaces and positions) to be assigned over
        and over again, thus making it appear as if the
        animation is paused).

        To make the animations play normally, you have to
        manually call self.play after instantiating this
        class.
        """
        ### alias "steps"
        n = steps

        ### pick node names list, with preference to the
        ### one provided by the user, if present
        node_names_list = nodes or self.updating_order

        ### iterate over each node listed, rotating the
        ### timing indices and updating the respective
        ### values using the current timing index

        for node_name in node_names_list:

            ### retrieve values and timing for the named node
            ### in the current animation

            node_values = self.values[self.anim_name][node_name]
            node_timing = self.timing[self.anim_name][node_name]

            ### rotate the time indices

            node_timing["surface_indices"].walk(n)
            node_timing["position_indices"].walk(n)

            ### retrieve the current time index (the first
            ### one) for both surfaces and positions

            surf_index = node_timing["surface_indices"][0]
            pos_index = node_timing["position_indices"][0]

            ### finally, using the indices just retrieved,
            ### store current surface reference and set
            ### position of the node

            ## reference node
            node = self.node_map[node_name]

            ## assign surface reference to image attribute
            node.image = node_values["surfaces"][surf_index]

            ## set position using related method
            node.set_pos(node_values["positions"][pos_index])

    step_forth = partialmethod(update_nodes, 1)
    two_steps_forth = partialmethod(update_nodes, 2)

    step_back = partialmethod(update_nodes, -1)
    two_steps_back = partialmethod(update_nodes, -2)

    ### drawing routines

    # obs.: the get_nodes method from the Utilities
    # class extension isn't used here (map is used instead)
    # for the sake of efficiency, since get_nodes perform
    # some extra checks which aren't needed here;

    def draw_nodes(self):
        """Draw each node."""
        for node in map(self.node_map.__getitem__, self.drawing_order):
            node.draw()

    def draw_nodes_and_bounding_rect(self):
        """Draw each node and its rects."""
        for node in map(self.node_map.__getitem__, self.drawing_order):

            node.draw()
            draw_rect(SCREEN, ORANGE, node.rect, 2)

    def draw_nodes_and_col_rect(self):
        """Draw each node and its collision rect, if any."""
        for node in map(self.node_map.__getitem__, self.drawing_order):

            node.draw()

            try:
                draw_rect(SCREEN, RED, node.col_rect, 2)

            # node.col_rect is None instead of Rect
            except TypeError:
                pass



"""Facility for cloud classes.

Used as parallax decorative elements.
"""

from random import seed, randint
from functools import partial

from ..appcommon.anim.player.main import AnimationPlayer
from ..appcommon.task import add_task


MOVE_TIME = 9000
ANIM_TIME = 12000


class Cloud:
    """A simple cloud."""

    def __init__(self, prop_name, bottomleft, anim_name):
        """Position and perform setups.

        prop_name
            String representing the name of the prop.
        bottomleft
            Bottomleft coordinates for a pygame.Rect instance.
        anim_name
            String representing name of the animation to
            use.
        """
        self.prop_name = prop_name
        self.anim_player = AnimationPlayer(
            self, prop_name, coordinates_name="bottomleft", coordinates_value=bottomleft
        )

        self.anim_player.switch_animation(anim_name)

        move_task = partial(self.rect.move_ip, 1, 1)
        add_task(move_task, MOVE_TIME, cyclic=True)

        anim_task = self.anim_player.step_forth

        seed()

        # XXX it is better to move only once by the
        # amount you pick, instead of multiple times
        # moving single units
        for i in range(randint(0, 15)):
            anim_task()

        add_task(anim_task, ANIM_TIME, cyclic=True)

    def scroll(self, dx, dy):
        """Scroll by dx and dy in respective axis.

        dx
            Integer indicating scrolling in x axis.
        dy
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(dx / 80, dy / 80)

    def update(self):
        """Update animation."""
        self.anim_player.update()

    def draw(self):
        """Draw animation."""
        self.anim_player.draw()



### third-party import
from pygame.time import set_timer

### local imports

from ...logconfig import APP_LOGGER

from ...config import GAME_REFS, SWITCH_LEVEL_TRIGGER


def set_level_switching(after, level_name, destination=None):
    """Set a level switching event. Store level name, too.

    after (integer)
        number of milliseconds after which to generate
        the event.
    level_name
        the basename of the next level to be loaded.
    """
    GAME_REFS.next_level = level_name
    if destination:
        GAME_REFS.destination = destination

    milliseconds = after
    set_timer(SWITCH_LEVEL_TRIGGER, milliseconds)



"""Facility to manage fonts."""

from os.path import join

### third-party imports

from pygame.font import Font
from pygame.font import init as init_font

### local imports

from .config import FONTS_DIR, FONTS_META_FILE
from .common.jsonhandler import load_json

init_font()

### create and process font path map

FONT_PATH_MAP = {
    "default": "source_sans_pro_regular.ttf",
}

for key, font_name in FONT_PATH_MAP.items():
    FONT_PATH_MAP[key] = join(FONTS_DIR, font_name)

### create a CHAR_CODE_MAP

## load fonts metadata
FONTS_META_MAP = load_json(FONTS_META_FILE)

## create map

CHAR_CODE_MAP = {
    font_key: FONTS_META_MAP[font_key] for font_key in FONT_PATH_MAP.keys()
}


### create a map to store commonly used font objects;
### that is, whenever a font object is created, it is stored
### for future usage, so no instantiation is required twice

FONT_MAP = {}

### utility function for general usage: getting fonts


def get_font(size, font_style="default"):
    """Return default font of specified size."""
    ### store arguments in tuple to use as a dictionary key
    key = (font_style, size)

    ### try returning font from font map using the key
    try:
        return FONT_MAP[key]

    ### otherwise create, store and return the new font obj
    except KeyError:

        ### retrieve font path
        font_path = FONT_PATH_MAP[font_style]

        ### instantiate font object
        font = Font(font_path, size)

        ### store it in the font map and return it

        FONT_MAP[key] = font
        return font



"""Facility for the rocket scorpion actor class."""

from ..appcommon.anim.player.main import AnimationPlayer


class RocketScorpion(object):
    """A stinger launcher scorpion."""

    level = None

    def __init__(self, prop_name, coordinates):
        """Initialize superclass and variables.

        prop_name
            A string representing the prop name.
        coordinates
            List or tuple of x and y coordinates representing
            the desired bottomleft coordinates for the
            rect attribute.
        """
        self.prop_name = prop_name

        self.anim_player = AnimationPlayer(
            self,
            prop_name,
            coordinates_name="bottomleft",
            coordinates_value=coordinates,
        )

        self.anim_player.play()

    def update(self):
        """Update animation."""
        self.anim_player.update()

    def draw(self):
        """Draw animation."""
        self.anim_player.draw()

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)



"""Facility for simple object display."""

from functools import partialmethod

### local imports

from ..config import SCREEN_RECT

from ..common.behaviour import empty_function

from ..appcommon.surf import INVISIBLE_SURF, render_image
from ..appcommon.autoblit import BasicObject

from .proc import process_tasks


class SimpleObject(BasicObject):
    """Simple object to hold sprite."""

    def __init__(
        self, image_name, tasks, exchange_pos_attrs=("center", "center"), offset=(0, 0)
    ):
        """Create attributes."""
        ### store a loaded image attribute and also
        ### reference it on the self.image attribute

        self.image = self.loaded_image = render_image(image_name)

        ### create, store and position a pygame.Rect
        ### instance from the surface referenced on the
        ### image attribute

        ## create and store rect
        self.rect = self.image.get_rect()

        ## position it

        rect_attr, screen_rect_attr = exchange_pos_attrs

        screen_pos = getattr(SCREEN_RECT, screen_rect_attr)

        setattr(self.rect, rect_attr, screen_pos)

        # offset if requested
        self.rect.move_ip(offset)

        self.visibility_off()

        ### define update behaviour
        self.update = empty_function

        ### process tasks
        process_tasks(self, tasks)

    def change_visibility(self, visible=True):
        """Turn the object invisible."""
        if visible:
            self.image = self.loaded_image
        else:
            self.image = INVISIBLE_SURF

    visibility_on = partialmethod(change_visibility, True)
    visibility_off = partialmethod(change_visibility, False)



"""Facility for common collision operations.

Those collision operations are usually abstractions of the
pygame.Rect class underlying methods.
"""

from ..config import SCREEN_RECT
from .autoblit import BlitterSet


# XXX
# Lots of changes to be expected here, specially
# regarding collisions

# XXX
# Proposal about collision handling:
# Each object would update its own collisions by checking,
# for instance, if an specific damaging rect inflicted
# damage on another object on the screen by asking said
# object whether it has a damageable rect colliding with
# the damaging rect. It could be flexible enough to
# allow for multiple kinds of collisions using rects, like
# checking for rect collisions or a radius reach, etc, etc.
# (I could describe it better)
# Ponder.


def does_touch_screen(obj):
    """Return True if obj touches screen.

    obj
        Any gaming obj.
    """
    if SCREEN_RECT.colliderect(obj.rect):
        return True
    else:
        return False


def get_objs_onscreen(group):
    """Return all objs which touch the screen.

    group
        A appcommon.autoblit.BlitterSet object.
    """
    onscreen = BlitterSet()
    for obj in group:
        if does_touch_screen(obj):
            onscreen.add(obj)
    return onscreen



"""Facility for common behaviour.


### definitions for doctests

>>> from functools   import partial
>>> from collections import deque

"""

from functools import partial, reduce
from operator import methodcaller
from collections import deque


class CallList(list):
    """A list of callable objects to be executed in order."""

    def __init__(self, *args, **kwargs):
        """Check callable state of items.

        Extends list.__init__
        """
        super().__init__(*args, **kwargs)
        if not all(map(callable, self)):
            raise ValueError("All items must be callable")

    def __call__(self):
        """Call all items."""
        for item in self:
            item()

    def append(self, item):
        """Append item if callable. Extends list.append."""
        if not callable(item):
            raise ValueError("All items must be callable")
        else:
            super().append(item)

    def insert(self, index, item):
        """Insert item if callable. Extends list.insert."""
        if not callable(item):
            raise ValueError("All items must be callable")
        else:
            super().insert(index, item)

    def extend(self, iterable):
        """Extend call list if all in iterable are callable.

        Extends list.extend.
        """
        if not all(map(callable, iterable)):
            raise ValueError("All items must be callable")
        else:
            super().extend(iterable)


class NestedPartial(partial):
    """A partial which retrieves return values from args.

    Instances of this class can be nested in a tree like
    structure. Useful to stack the execution of multiple
    functions when their return values serve as arguments
    for each other.

    >>> def get_value(value): return value

    >>> get_me  = partial(get_value,  'me')
    >>> get_you = partial(get_value, 'you')
    >>> get_we  = partial(get_value,  'we')

    >>> print_me = NestedPartial(
    ...            print, get_me, get_you, end=get_we)
    >>> print_me()
    me youwe

    Notice in the doctest above that since get_me and
    get_wehee functions are in the end of the 'tree-like'
    nesting, they don't need the extra functionality of
    a NestedPartial and can be partials or any other
    callable (as long as you don't need to provided extra
    arguments, in which case partials are required).
    """

    def __call__(self):
        """Return func return value after executing args."""
        ### retrieve positional arguments' return values if
        ### they are callable, otherwise just use themselves
        args = (arg() if callable(arg) else arg for arg in self.args)

        ### retrieve keyword arguments' return values if they
        ### are callable, otherwise just use themselves
        kwargs = {
            key: value() if callable(value) else value
            for key, value in self.keywords.items()
        }

        ### return the return value of self.func called
        ### with the resulting arguments
        return self.func(*args, **kwargs)


def empty_function():
    """Do nothing. An empty function.

    Replaces 'lambda:None' solution.

    Some respected members in the pygame community don't
    recommend the usage of lambda functions for the sake
    of clarity. I'm still divided about it, but for such
    simple function declaration there's no cost at all
    to declare and keep it and when it is imported it has
    the advantage of communicating that some behaviour in
    the module has a 'switched off' state.

    Profiling have shown a very tiny and practically
    insignificant speed gain (less than fractions of
    microseconds) over the lambda solution, though not
    consistent through many tests.

    Try yourself on your command line:

    python -m timeit --setup "f = lambda:None" "f()"
    python -m timeit --setup "def f():pass"    "f()"

    Doctest of my life

    >>> empty_function()
    >>>

    Disclaimer: for a function which does nothing, I know
    it has quite the docstring (including its doctest),
    but the discussion and design considerations are always
    important; you don't need to be as detailed as I am,
    but reviewing your design decisions and documenting them
    is a very constructive habit and helps to develop your
    analytical thinking.
    """


def get_nested_value(d, *strings):
    """Return inner value from nested dict.

    d (dict)
        Dictionary containing other nested dictionaries.
    strings (string arguments)
        Strings used to retrieve innermost item.

    >>> d = {
    ...   "john" : {
    ...     "id" : 22,
    ...     "school" : {
    ...       "id" : 1043,
    ...       "name" : "St. Lewis"
    ...     }
    ...   }
    ... }
    >>> get_nested_value(d, "john", "school", "id")
    1043

    ### curiosity

    A higher level version, could work with any nested
    objects with a __getitem__ method,
    and instead of strings could receive any arbitrary
    value, as long as it could be used by __getitem__ to
    retrieve the next one

    get_nested_value(obj, values):
        for item in values: obj = obj[item]
        return obj

    this would allows us to retrieve deep nested values
    within dicts, lists, tuples, etc.

    Maybe I should profile this with the higher level version
    to see if it is worth to use the more general approach
    of the higher level version.
    """
    return reduce(dict.__getitem__, strings, d)


### attribute utilities

## routine setting


def get_routine_caller(obj, routine_name, starting_value=empty_function):
    """Return function which calls an specific routine.

    Also sets an initial value for the routine in the
    attribute.

    obj (any object which can have attributes set on it)
    routine_name (string)
        name of the attribute where to store the routine.
    starting_value (callable)
        any callable to be used as the starting value of the
        routine. If no value is passed, the empty_function
        is called instead.

    >>> class Obj: pass
    ...
    >>> obj = Obj()
    >>> def a(): print("a")
    ...
    >>> def b(): print("b")
    ...
    >>> routine_caller = get_routine_caller(obj, "printer")
    >>> routine_caller()
    >>> obj.printer = a
    >>> routine_caller()
    a
    >>> obj.printer = b
    >>> routine_caller()
    b
    >>>

    """
    setattr(obj, routine_name, starting_value)

    return partial(methodcaller(routine_name), obj)


## toggling utilities


def get_attribute_rotator(obj, attr_name, values):
    """Return a function to rotate obj attribute values.

    Creates a partial from the provided arguments and the
    rotate_attribute function.

    obj (any object which can have attributes set on it)
        object whose attribute is used to hold the value.
    attr_name (string)
        name of the attribute where to store a value.
    values (iterable containing any python objects/values)
        values between which to toggle.

    Also sets the first value on the obj attribute.

    >>> values = [12, 32, 45, 26]
    >>> rotate = get_attribute_rotator
    """
    values_deque = deque(values)

    setattr(obj, attr_name, values_deque[0])

    return partial(rotate_attribute, obj, attr_name, values_deque)


def rotate_attribute(obj, attr_name, values_deque):
    """Rotate attribute value between those provided.

    This function is meant to be turned into partials
    by the get_attribute_rotator function for more efficient
    usage.

    Tip: passing a deque with two values creates a simple
    very useful toggle.

    obj (any object which can have attributes set on it)
        object whose attribute is used to hold the value.
    attr_name (string)
        name of the attribute where to store the value.
    values_deque (deque containing any python object/values)
        values between which to toggle.

    >>> class Obj: pass
    ...
    >>> obj = Obj()
    >>> d = deque(["a", "b", "c"])
    >>> rotate = partial(rotate_attribute, obj, "char", d)

    >>> rotate()
    >>> obj.char
    'b'
    >>> rotate()
    >>> obj.char
    'c'

    >>> obj.char = "b"
    >>> rotate()
    >>> obj.char
    'a'

    Notice in the last doctest section that even when we
    set the char attribute to "b" the next value of char
    after rotating is "a". This is because the assignement
    follows the rotation of the deque, without taking into
    consideration the current value.

    If you need a function that takes into account the
    current value before assigning a new one you should use
    the get attribute toggler function.
    """
    values_deque.rotate(-1)
    setattr(obj, attr_name, values_deque[0])


def get_attribute_toggler(
    obj, attr_name, starting_value, mapping_or_value, default=None
):
    """Return function to toggle obj atribute values.

    Creates a partial from the provided arguments using
    the toggle_attribute function.

    obj (any object which can have attributes set on it)
        object whose attribute is used to hold the value.
    attr_name (string)
        name of the attribute where to store a value.
    starting_value (any hashable obj/value),
        starting value to be assigned to the attribute.

    mapping_or_value (a dict or any hashable obj/value)
        The type of mapping_or_value may be different depending
        on two scenarios:

        1) If you have more than two values between which
           you desired to toggle though, you must then
           provide a mapping which associates the current
           value of the attribute with the desired value to
           be assigned when toggling.


        2) value to which we should toggle when executing
           the partial returned. Executing the partial
           returned again should set the previous value
           in the attribute again and so on.

        Either way the arguments are stored in a mapping
        so they fit the first scenario, thus complying with
        the underlying toggle_attribute function.

        The first scenario is just a simplification, since
        having two values to toggle would mean you want to
        toggle between them in this context. There's no
        absolute need to make each value reference each other,
        though. The work of the toggle is just to set a
        value for the attribute based on the current value
        stored there. If no value is found in the dict which
        is associated with the current value, the toggler
        (partial) will do nothing.
    default (any python value/object)
        Value to use only in case no key is found.

    >>> class Obj: pass
    ...
    >>> obj = Obj()
    >>> toggle_char = get_attribute_toggler(
    ...                   obj, 'char', 'a', 'b')
    >>> obj.char
    'a'
    >>> toggle_char()
    >>> obj.char
    'b'
    >>> obj.char = 'c'
    >>> toggle_char()
    >>> obj.char # without default it returns None
    >>>

    >>> obj.char = 'a'
    >>> toggle_char()
    >>> obj.char
    'b'
    >>> d = {1:5, 5:22, 22:13, 13:1}
    >>> toggle_number = get_attribute_toggler(
    ...                     obj, 'number', 22, d, 100)
    >>> obj.number
    22
    >>> toggle_number()
    >>> obj.number
    13
    >>> toggle_number()
    >>> obj.number
    1
    >>> obj.number = 342
    >>> toggle_number()
    >>> obj.number # 342 isn't a key, so returns default
    100

    """
    ### set the starting value of the attribute
    setattr(obj, attr_name, starting_value)

    ### now check the type of the arguments so you can
    ### create or reference the cross_mapped_values dict

    ## if mapping_or_value is a dict, reference it
    if isinstance(mapping_or_value, dict):
        cross_mapped_values = mapping_or_value

    ## otherwise, mapping_or_value is actually a value,
    ## so just use it to build a new dict
    else:
        another_value = mapping_or_value

        cross_mapped_values = {
            starting_value: another_value,
            another_value: starting_value,
        }

    ### return function
    return partial(toggle_attribute, obj, attr_name, cross_mapped_values, default)


def toggle_attribute(obj, attr_name, cross_mapped_values, default=None):
    """Toggle attribute between values a and b.

    This function is meant to be turned into partials
    by the get_attribute_toggler function for more efficient
    usage.

    It is slightly different from the toggler achievable with
    the get_attribute_rotator function because it checks the
    value of the attribute before toggling, so there is no
    intrinsic order. The value set depends solely on the
    current value.

    obj (any object which can have attributes set on it)
        object whose attribute is used to hold the value.
    attr_name (string)
        name of the attribute where to store the value.
    cross_mapped_values (dict)
        its keys reference each other. Such keys represent
        values from which and to which to toggle.
    default (any python value/object)
        Value to use only in case no key is found.

    >>> class Obj: pass
    ...
    >>> obj = Obj()
    >>> d = {"a":"b", "b":"a"}
    >>> toggler = partial(toggle_attribute, obj, "char", d)

    >>> obj.char = "a"
    >>> toggler()
    >>> obj.char
    'b'
    >>> toggler()
    >>> obj.char
    'a'
    """
    ### retrieve current attribute value to use as key
    key = getattr(obj, attr_name)

    ### retrieve value using key, also defining a default
    ### to be returned if needed
    value = cross_mapped_values.get(key, default)

    ### assign value to the attribute
    setattr(obj, attr_name, value)



"""Facility for the inventory manager."""

from collections import deque

### local import
from ..throwing.mushroom import Mushroom


# XXX
# when coding the item switching functionality,
# remember to also make the needed changes
# on the equipped display class.


def instantiate_item(item_name, player):
    """Instantiate item based on name.

    item_name
        A string representing the name of an equippable
        item. It is used to instantiate it.
    player
        A reference for an instance of the
        player.main.Player class, from which to harvest
        info.
    """
    if item_name == "mushroom":
        prop_name = item_name
        coordinates = [0, 0]
        level = player.level
        item = Mushroom(prop_name, coordinates, level)

    return item


class InventoryManager:
    """An inventory manager."""

    def __init__(self, state_data, equipped_display, player):
        """Initialize object.

        state_data
            json data loaded from the slot .stt file.
        equipped_display
            A reference to a
            player.equippeddisplay.EquippedDisplay instance
            used to communicate changes on the inventory.
        player
            A reference to a player.main.Player
            instance.
        """
        self.state_data = state_data
        self.equipped_display = equipped_display
        self.player = player

        equipped_item = self.state_data.get("equipped_item", "mushroom")
        self.equipped_item_name = equipped_item

        self.prepare_items()
        self.update_equipped_display()

    def prepare_items(self):
        """Instantiate all present items."""
        self.item_name_deque = deque()
        self.item_objects = {}

        inventory = self.state_data.get("inventory")
        if inventory:
            for item_name in inventory:
                self.item_name_deque.append(item_name)
                self.item_objects[item_name] = []

                amount = inventory.get(item_name, 0)
                for n in range(amount):
                    obj = instantiate_item(item_name, self.player)
                    self.item_objects[item_name].append(obj)

    def update_equipped_display(self):
        """Send item and count data to equipped display."""
        inventory = self.state_data.get("inventory")
        if inventory:
            self.equipped_item_count = inventory[self.equipped_item_name]

            self.equipped_display.update_visual_data(
                self.equipped_item_name, self.equipped_item_count
            )

    def add(self, item):
        """Return True if item is added.

        item
            A instance item to be added in the inventory.
        """
        item_name = item.prop_name

        try:
            amount = self.state_data["inventory"][item_name]

        except KeyError:

            if not self.state_data.get("inventory"):
                self.state_data["inventory"] = {}

            self.state_data["inventory"][item_name] = 1

            self.item_objects[item_name] = []
            self.item_objects[item_name].append(item)

            if item_name not in self.item_name_deque:
                self.item_name_deque.append(item_name)

            self.update_equipped_display()
            return True

        else:

            if amount < 6:
                self.state_data["inventory"][item_name] += 1

                self.item_objects[item_name].append(item)

                self.update_equipped_display()
                return True

            else:
                self.player.say_full_inventory()
                return False

    def request_equipped_item(self):
        """If manage to remove item, pass it to player."""
        item_name = self.equipped_item_name

        if self.remove(item_name):
            item = self.item_objects[item_name].pop()

        else:
            item = None

        return item

    def remove(self, item_name):
        """Return True if succesfully remove item."""
        try:
            amount = self.state_data["inventory"][item_name]
        except KeyError:
            return False
        else:
            if amount >= 1:
                self.state_data["inventory"][item_name] += -1
                self.update_equipped_display()
                return True
            else:
                return False



"""Facility for sign surface generation."""

from pygame import Rect
from pygame.draw import rect as draw_rect, polygon as draw_polygon

from ..palette import BLACK, ARBITRARY_COLORKEY, SIGN_POLE_COLOR, has_same_rgb
from .surf import render_rect
from .text.main import render_text


# XXX
# art: make signs just a bit more smaller,
# to be more proportional to the player
# and npcs bodies.

# XXX
# Make the box adjust height too,
# so you're able to change font size
# and the box would adjust  accordingly.
# Use a fixture for that. Then maybe
# feature toggle. Signs could use surfaces
# too.


def sign_factory(text, fg_color, bg_color, signpost_height=0, direction=None):
    """Return a sign or signpost surface.

    text
        A string representing the message on the sign.

    fg_color
    bg_color
        A tuple or list of r, g, b values which are
        integers ranging from 0 to 255. This color will
        be in several pygame.draw drawing calls.

        fg_color defines the foreground color while
        bg_color defines the background color.

        Though this function wasn't intended to be used
        in this way an optional alpha value within the
        same range can also be provided in the tuple
        or list, and the purpose of this can be read
        in the link bellow:
        https://www.pygame.org/docs/ref/draw.html

        Here's the specific excerpt:
        "Most of the arguments accept a color argument
        that is an RGB triplet. These can also accept
        an RGBA quadruplet. The alpha value will be
        written directly into the Surface if it contains
        pixel alphas, but the draw function will not draw
        transparently. The color argument can also be an
        integer pixel value that is already mapped to the
        Surface's pixel format."

    signpost_height (defaults to 0)
        An integer describing the height of a pole to
        support the sign, if you want one. If not provided
        or 0 (the default), the sign won't have a post and
        it will be as if it were put on the wall.

    direction (defaults to None)
        If None the sign will be a rectangle. If one
        of ('n', 's', 'e' or 'w') is provided than
        the sign will be an arrow pointing in the
        specified direction. 'n' stands for north,
        's' for south, 'e' for east and 'w' for west.
    """

    # about function below:
    #
    # generated surfaces take this final form:
    #
    #    rectangular,            arrow-shaped,
    #        no post                  no post
    #                                             |\
    #     _______________      ___________________| \
    #    |               |    |                      \
    #    |               |    |                       \
    #    | Factory exit  |    | City entrance ahead    .
    #    |               |    |                       /
    #    |_______________|    |____________________  /
    #                                             | /
    #                                             |/
    #
    #     rectangular,           arrow-shaped,
    #        has post                has post     |\
    #     _______________      ___________________| \
    #    |               |    |                      \
    #    |               |    |                       \
    #    | Fancy lodging |    | Frozen waterfalls II   .
    #    |               |    |                       /
    #    |_______________|    |____________________  /
    #           |  |                   |  |       | /
    #           |  |                   |  |       |/
    #           |  |                   |  |
    #           |__|                   |__|
    #
    #    Notice the length is automatically adjusted to
    #    comprise the text surface. The sign can take a
    #    rectangular form or be shaped like an arrow pointing
    #    into specific directions. The sign can have an optional
    #    post.

    if has_same_rgb(fg_color, ARBITRARY_COLORKEY):
        msg = "'fg_color' can't be equal to" + " 'ARBITRARY_COLORKEY' constant."
        raise ValueError(msg)
    elif has_same_rgb(bg_color, ARBITRARY_COLORKEY):
        msg = "'bg_color' can't be equal to" + " 'ARBITRARY_COLORKEY' constant."
        raise ValueError(msg)

    directions = ("n", "s", "e", "w")
    if direction not in directions:
        shape = "rectangular"
    else:
        shape = "arrow"

    text_surf = render_text(text, font_size=22, foreground_color=fg_color, padding=0)

    text_width = text_surf.get_size()[0]

    if shape == "rectangular":
        sign_surf = render_rect(text_width + 20, 35)
        sign_surf.fill(bg_color)

        outline_rect = sign_surf.get_rect()
        outline_rect.width += -1
        outline_rect.height += -1
        draw_rect(sign_surf, BLACK, outline_rect, 4)

        sign_surf.blit(text_surf, (10, 3))
        sign_surf_width, sign_surf_height = sign_surf.get_size()

        if not signpost_height:
            offset = 5
            holder_surf = render_rect(
                sign_surf_width + offset, sign_surf_height + offset
            )
            holder_surf.fill(ARBITRARY_COLORKEY)

            shadow_rect = sign_surf.get_rect()
            shadow_rect.x += offset
            shadow_rect.y += offset
            draw_rect(holder_surf, BLACK, shadow_rect, 0)

            holder_surf.blit(sign_surf, (0, 0))

            # this eliminates an artifact
            final_touch_rect = Rect((sign_surf_width - 2, sign_surf_height - 2), (4, 4))
            draw_rect(holder_surf, BLACK, final_touch_rect, 0)

        elif signpost_height:
            height_offset = sign_surf_height + 20
            signpost_offset_height = signpost_height + height_offset
            pole_width = 18
            pole_surf = render_rect(pole_width, signpost_offset_height)
            pole_surf.fill(SIGN_POLE_COLOR)

            pole_outline_rect = pole_surf.get_rect()
            pole_outline_rect.width += -1
            pole_outline_rect.height += -1
            draw_rect(pole_surf, BLACK, pole_outline_rect, 4)

            triangle = [
                (0, height_offset),
                (40, height_offset),
                (40, height_offset + 40),
            ]
            draw_polygon(pole_surf, BLACK, triangle, 0)

            holder_surf = render_rect(sign_surf_width, signpost_offset_height)
            holder_surf.fill(ARBITRARY_COLORKEY)

            pole_x = round(sign_surf_width / 2) - round(pole_width / 2)
            holder_surf.blit(pole_surf, (pole_x, 0))
            holder_surf.blit(sign_surf, (0, 20))

    elif shape == "arrow":
        # arrow and its points
        #             . (f)
        #  (a)        |\
        #  ._ _ _ _ _.| \
        #  |        (g)  \
        #  |              \
        #  |               . (e)
        #  |              /
        #  ._ _ _ _ _ .  /
        #  (b)     (c)| /
        #             |/
        #             . (d)

        a, b = (0, 10), (0, 40)

        padx = 10
        padx_text_width = text_width + (padx * 2)
        c, g = ((padx_text_width, b[1]), (padx_text_width, a[1]))
        f, d = (g[0], 0), (c[0], c[1] + 10)

        x_distance_e_to_d = 25
        e = (c[0] + x_distance_e_to_d, a[1] + 15)

        arrow = [a, b, c, d, e, f, g]

        sign_surf = render_rect(e[0], d[1])
        sign_surf.fill(ARBITRARY_COLORKEY)

        draw_polygon(sign_surf, bg_color, arrow, 0)
        draw_polygon(sign_surf, BLACK, arrow, 3)

        if not signpost_height:
            shadow_surf = render_rect(*sign_surf.get_size())
            shadow_surf.fill(ARBITRARY_COLORKEY)
            draw_polygon(shadow_surf, BLACK, arrow, 0)
            shadow_surf.set_colorkey(ARBITRARY_COLORKEY)

        if direction == "e":
            sign_surf.blit(text_surf, (10, 10))

            if not signpost_height:
                offset = 5
                temp_width, temp_height = sign_surf.get_size()
                holder_surf = render_rect(temp_width + offset, temp_height + offset)
                holder_surf.fill(ARBITRARY_COLORKEY)
                sign_surf.set_colorkey(ARBITRARY_COLORKEY)
                holder_surf.blit(shadow_surf, (offset, offset))
                holder_surf.blit(sign_surf, (0, 0))

        elif direction == "w":
            sign_surf = flip(sign_surf, True, False)
            sign_surf.blit(text_surf, (60, 10))

            if not signpost_height:
                shadow_surf = flip(shadow_surf, True, False)
                offset = 5
                temp_width, temp_height = sign_surf.get_size()
                holder_surf = render_rect(temp_width + offset, temp_height + offset)
                holder_surf.fill(ARBITRARY_COLORKEY)
                sign_surf.set_colorkey(ARBITRARY_COLORKEY)
                holder_surf.blit(shadow_surf, (offset, offset))
                holder_surf.blit(sign_surf, (0, 0))

        elif direction in ("n", "s"):
            angle = 90 if direction == "n" else -90
            sign_surf.blit(text_surf, (10, 10))
            sign_surf = rotate(sign_surf, angle)

            if not signpost_height:
                shadow_surf = rotate(shadow_surf, angle)
                offset = 5
                temp_width, temp_height = sign_surf.get_size()
                holder_surf = render_rect(temp_width + offset, temp_height + offset)
                holder_surf.fill(ARBITRARY_COLORKEY)
                sign_surf.set_colorkey(ARBITRARY_COLORKEY)
                holder_surf.blit(shadow_surf, (offset, offset))
                holder_surf.blit(sign_surf, (0, 0))

        if signpost_height:
            sign_surf_width, sign_surf_height = sign_surf.get_size()
            height_offset = sign_surf_height
            signpost_offset_height = signpost_height + height_offset

            pole_width = 22
            pole_surf = render_rect(pole_width, signpost_offset_height)
            pole_surf.fill(SIGN_POLE_COLOR)

            pole_outline_rect = pole_surf.get_rect()
            pole_outline_rect.width += -1
            pole_outline_rect.height += -1

            # triangle used as a shadow cast on pole
            # just below the sign
            if direction in ("e", "w"):
                triangle = [
                    (0, height_offset - 10),
                    (40, height_offset - 10),
                    (40, height_offset + 30),
                ]
            elif direction == "n":
                triangle = [
                    (0, height_offset + 5),
                    (40, height_offset + 5),
                    (40, height_offset + 45),
                ]
            else:
                triangle = [
                    (0, height_offset - 25),
                    (40, height_offset - 25),
                    (40, height_offset + 15),
                ]

            draw_rect(pole_surf, BLACK, pole_outline_rect, 4)
            draw_polygon(pole_surf, BLACK, triangle, 0)

            holder_surf = render_rect(sign_surf_width, signpost_offset_height)
            holder_surf.fill(ARBITRARY_COLORKEY)

            pole_x = round(sign_surf_width / 2) - round(pole_width / 2)
            if direction == "e":
                pole_x += -10
            elif direction == "w":
                pole_x += 10

            holder_surf.blit(pole_surf, (pole_x, 0))
            sign_surf.set_colorkey(ARBITRARY_COLORKEY)
            holder_surf.blit(sign_surf, (0, 5))

    holder_surf.set_colorkey(ARBITRARY_COLORKEY)
    sign_surf = holder_surf

    return sign_surf



"""Custom exceptions."""


class ManagerSwitchException(Exception):
    """Raised when the update manager is switched.

    The update manager is any object whose reference
    is stored in the local variable of the gameloop.run_game
    function.
    """

    def __init__(self, update_manager_name):
        """Store the new update manager."""
        self.update_manager_name = update_manager_name
        super().__init__("Switching update manager.")


class LevelSwitchException(ManagerSwitchException):
    """Raised when switching between levels."""


class RestartLevelException(Exception):
    """Raise when player dies to restart level."""


class QuitGameException(Exception):
    """Raised to quit the game loop."""



"""Facility for scene playing."""

from os.path import join

### third-party imports

from pygame import QUIT, KEYDOWN, K_SPACE

from pygame.display import update
from pygame.event import get as get_events
from pygame.time import set_timer

### local imports

from ..config import GAME_REFS, SCENES_DIR, SCREEN, SCREEN_RECT, SWITCH_LEVEL_TRIGGER

from ..common.jsonhandler import load_json

from ..appcommon.task import update_task_manager, clear_task_manager
from ..appcommon.surf import render_rect
from ..appcommon.dialog import clear_messages
from ..appcommon.autoblit import BlitterSet
from ..appcommon.exception import QuitGameException, ManagerSwitchException

from ..appcommon.behaviour.switch import set_level_switching

from ..music import MUSIC_MANAGER
from ..palette import BLACK

from .simple import SimpleObject
from .animated import AnimatedObject

from ..logconfig import APP_LOGGER

logger = APP_LOGGER.getChild(__name__)


class ScenePlayer:
    """An update manager which displays timed media.

    This is a generic update manager for sections of the
    game which only require showing and hiding media in given
    intervals (for instance, logo screen, etc.).
    """

    def __init__(self, scene_filename):
        """Assign variables and perform setups.

        scene_filename (string)
            Used to retrieve data about the scene.
        """
        scene_path = join(SCENES_DIR, scene_filename)
        self.scene_data = load_json(scene_path)

        self.set_scene()

    def set_scene(self):
        """Set flags, create objects, set events."""
        ### set flag to (dis)allow skipping scene
        self.can_skip = self.scene_data.get("can_skip")

        ### set music

        mood = self.scene_data.get("mood", "casual")
        MUSIC_MANAGER.cue(mood)

        ### set background

        background_color = self.scene_data.get("background_color", BLACK)

        self.background = render_rect(*SCREEN_RECT.size, background_color)

        ### set objects
        self.set_objects()

        ### set events
        self.set_events()

    def set_objects(self):
        """Set objects."""
        self.objects = BlitterSet()

        for obj_data in self.scene_data["objects"]:

            ### get obj type
            obj_type = obj_data["obj_type"]

            ### store condition
            is_simple = obj_type.lower() == "simple"

            ### pick proper class according to condition
            instantiator = SimpleObject if is_simple else AnimatedObject

            ### retrieve keyword arguments
            kwargs = obj_data["kwargs"]

            ### instantiate and add object

            obj = instantiator(**kwargs)

            self.objects.add(obj)

    def set_events(self):
        """Set events on a timer."""

        # TODO instead of calling set_level_switching
        # here, which uses milliseconds, create a
        # partial with the values and add it as a
        # task in the task manager (with an small delay
        # of half a second, maybe)
        # This is better than relying on the milliseconds
        # count which might be off if the CPU delays
        # execution (for instance, due to many apps
        # open at the same time); the time in milliseconds,
        # when present, could be converted into frames and
        # be added as a task the same way;

        for event in self.scene_data["events"]:

            if event["name"].lower() == "switch_level":

                milliseconds = event["milliseconds"]
                level_name = event["level_name"]

                set_level_switching(milliseconds, level_name)

    def control(self):
        """Update manager based on event_queue."""
        for event in get_events():

            if event.type == QUIT:
                raise QuitGameException

            elif event.type == SWITCH_LEVEL_TRIGGER:
                self.switch_level()

            elif event.type == KEYDOWN:
                if event.key == K_SPACE and self.can_skip:
                    self.switch_level()

    def update(self):
        """Update menu objects."""
        self.objects.update_objs()
        update_task_manager()

    def draw(self):
        """Draw objects and update screen."""
        SCREEN.blit(self.background, (0, 0))
        self.objects.draw_objs()

        update()

    def switch_level(self):
        """Switch to next level."""
        set_timer(SWITCH_LEVEL_TRIGGER, 0)

        clear_task_manager()
        clear_messages()

        level_name = GAME_REFS.next_level
        del GAME_REFS.next_level

        raise ManagerSwitchException(level_name)



from ...config import USER_SETTINGS, CONFIG_FILE

from ...common.jsonhandler import save_json

from ...music import MUSIC_MANAGER
from ...sound import set_sounds_volume


def apply_user_settings():
    """Apply user settings from USER_SETTINGS."""
    MUSIC_MANAGER.set_volume(USER_SETTINGS["music_volume"])
    set_sounds_volume(USER_SETTINGS["sound_volume"])
    save_json(USER_SETTINGS, CONFIG_FILE)



"""Player class facility."""

from types import SimpleNamespace
from random import seed, choice

# third-party imports
from pygame.time import set_timer

# local imports
from ..config import SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_RECT, RESTART_FROM_SAVE

from ..logconfig import APP_LOGGER

from ..common.math import get_straight_distance
from ..common.behaviour import CallList, empty_function

from ..appcommon.task import add_task
from ..appcommon.overrider import register_actions
from ..appcommon.dialoguefactory import dialoguebox_factory
from ..appcommon.anim.player.main import AnimationPlayer
from ..appcommon.behaviour.switch import set_level_switching

from ..sound import SOUNDS_MAP, set_sounds_volume

from .healthbar import HealthBar
from .dialoguemanager import DialogueManager
from .promptmanager import PromptManager
from .inventorymanager import InventoryManager
from .headsprite import Head
from .equippeddisplay import EquippedDisplay


logger = APP_LOGGER.getChild(__name__)


class Player:
    """The main character in the game: Dani."""

    def __init__(self, start_position, level, state_data, health=100):
        """Initialize superclass and variables."""
        self.anim_player = AnimationPlayer(
            self,
            "dani",
            coordinates_name="bottomleft",
            coordinates_value=start_position,
        )

        self.anim_player.play()

        self.level = level
        self.state_data = state_data
        self.health = health

        self.set_operation_structure()

    def set_operation_structure(self):
        """Set other important operational structures.

        Those include atributes, methods and other objects."""
        ### Instantiate assisting objects

        self.equipped_display = EquippedDisplay()
        self.inventory_manager = InventoryManager(
            self.state_data, self.equipped_display, self
        )

        self.healthbar = HealthBar(self)
        self.player_head = Head()

        ## Send relevant ones to heads-up display

        self.level.huds_group.update(
            [self.healthbar, self.player_head, self.equipped_display]
        )

        ### Define variables for control

        self.x_speed = 0
        self.x_accel = 1
        self.max_walk_speed = 10

        self.y_speed = 0
        self.gravity_accel = 2

        self.climb_speed = 5

        self.facing_right = True

        self.portal_holder = None

        ### Instantiate and setup additional managers

        self.dialogue_manager = DialogueManager(self)
        self.prompt_manager = PromptManager(self)

        ## Send additional managers to heads-up display

        self.level.huds_group.update([self.dialogue_manager, self.prompt_manager])

        dialogue = [
            {"player": "Oh, my bag don't support" " more than six of each item."}
        ]

        self.full_inventory_dialogue = dialoguebox_factory(dialogue)

        ### Set sounds

        seed()
        self.jump_sounds = [SOUNDS_MAP["player_jump01"], SOUNDS_MAP["player_jump02"]]

        ### Set movement routine (behaviour/physics)

        self.default_physics = CallList([self.move_x, self.move_y])

        self.climb_physics = CallList(
            [self.climb_x, self.climb_y, self.check_touching_ladder]
        )

        self.movement_routine = self.default_physics

        ### Set cooldown mechanisms
        self.ladder_to_ground_cooldown = SimpleNamespace(finished=True)

        ### send actions to control overrider to be
        ### referenced when needed
        self.map_actions()

    def update(self):
        """Update player states."""
        self.movement_routine()
        self.manage_animations()
        self.anim_player.update()

    def draw(self):
        """Draw animation."""
        self.anim_player.draw()

    def move_x(self):
        """Calculate and perform movement in x axis."""
        self.rect.x += self.x_speed

        self.check_handle_collision_x()
        self.check_block_collision_x()

    def move_y(self):
        """Calculate and perform movement in y axis."""
        self.apply_gravity_pull()
        self.rect.y += self.y_speed

        self.check_handle_collision_y()
        self.check_block_collision_y()

    def check_handle_collision_x(self):
        """Check for handle collision in x axis."""
        collided_handle = self.level.horizontal_barriers.collide_any(self)

        if collided_handle:
            self.warp_scroll_player(collided_handle)

    def check_handle_collision_y(self):
        """Check for handle collision in y axis."""
        collided_handle = self.level.vertical_barriers.collide_any(self)

        if collided_handle:
            self.warp_scroll_player(collided_handle)

    def check_block_collision_x(self):
        """Check for block collision in x axis."""
        collided_platform = self.level.onscreen_map["blocks"].collide_any(self)

        if collided_platform:
            is_going_right = self.x_speed > 0
            if is_going_right:
                self.rect.right = collided_platform.rect.left
            else:
                self.rect.left = collided_platform.rect.right
            self.x_speed = 0

    def check_block_collision_y(self):
        """Check for block collision in y axis."""
        collided_platform = self.level.onscreen_map["blocks"].collide_any(self)

        if collided_platform:
            # XXX
            # this test probably won't work though if a block
            # hits player's head. Maybe it's better to
            # determine if is hitting ground by the
            # difference between player's and block's topleft
            # rect position. Or maybe I'm just worrying too
            # much, since I don't plan to have blocks falling
            # on the stage.
            hit_ground = self.y_speed > 0

            if hit_ground:
                self.rect.bottom = collided_platform.rect.top

                try:
                    collided_platform.grass_walking(self, self.rect.midbottom)

                except AttributeError:
                    pass

            else:
                self.rect.top = collided_platform.rect.bottom

            self.y_speed = 0

    def climb_x(self):
        """Calculate and perform climbing in x axis."""
        if self.x_speed:
            self.rect.x += self.climb_speed if self.x_speed > 0 else -self.climb_speed

        self.check_handle_collision_x()
        self.check_block_collision_x()
        # must set to 0 only after block collision,
        # since collision mechanism in x axis rely on x_speed
        self.x_speed = 0

    def climb_y(self):
        """Calculate and perform climbing in y axis."""
        if self.y_speed:
            increment = self.climb_speed if self.y_speed > 0 else -self.climb_speed
            self.rect.y += increment
            if self.is_touching_ground():
                self.rect.y -= increment
                self.assign_default_physics()
                if self.ladder_to_ground_cooldown.finished:
                    self.ladder_to_ground_cooldown = add_task(empty_function, 100)

        self.check_handle_collision_y()
        self.check_block_collision_y()
        # must set to 0 only after block collision,
        # since collision mechanism in y axis rely on y_speed
        self.y_speed = 0

    def assign_climbing(self):
        """Check climbing state and assign if needed."""
        if self.movement_routine != self.climb_physics:
            self.movement_routine = self.climb_physics

    def assign_default_physics(self):
        """Check climbing state and assign if needed."""
        if self.movement_routine != self.default_physics:
            self.movement_routine = self.default_physics

    def check_touching_ladder(self):
        """Check if still touching ladders.

        Otherwise revert back to default physics.
        """
        if self.is_touching_ladder():
            pass
        else:
            self.assign_default_physics()

    def warp_scroll_player(self, handle):
        """Warp player according to scroll orientation.

        handle
            An sprite object whose collision is used to
            detect when to scroll the screen. Such
            scrolling is triggered by invoking a special
            method in the handle (but first we warp the
            player)."""
        if handle.rect.midtop == SCREEN_RECT.midtop:
            self.rect.y = SCREEN_HEIGHT - handle.rect.height - self.rect.height

        elif handle.rect.midbottom == SCREEN_RECT.midbottom:
            self.rect.y = handle.rect.height

        elif handle.rect.midright == SCREEN_RECT.midright:
            self.rect.x = handle.rect.width

        elif handle.rect.midleft == SCREEN_RECT.midleft:
            self.rect.x = SCREEN_WIDTH - handle.rect.width - self.rect.width
        handle.invoke()

    def apply_gravity_pull(self):
        """Move player according to gravity force."""
        will_traverse_floor = self.y_speed > self.rect.height

        if not will_traverse_floor:
            self.y_speed += self.gravity_accel

    ### animation management

    def manage_animations(self):
        """Change animation according to states."""
        if not self.is_touching_ground():

            if self.facing_right:
                self.anim_player.ensure_animation("jump")

            else:
                self.anim_player.ensure_animation("jump_reverse")
        else:
            if self.x_speed:

                if self.facing_right and self.x_speed > 0:
                    self.anim_player.ensure_animation("walk")

                elif not self.facing_right and self.x_speed < 0:
                    self.anim_player.ensure_animation("walk_reverse")

                elif self.facing_right and self.x_speed < 0:
                    self.anim_player.ensure_animation("decelerate_reverse")
                else:
                    self.anim_player.ensure_animation("decelerate")

            else:
                if self.is_crouched():
                    pass

                else:
                    if self.facing_right:
                        self.anim_player.ensure_animation("idle")
                    else:
                        self.anim_player.ensure_animation("idle_reverse")

    ### user controlled behaviour
    ### (there's more on interaction section when indicated)

    def go_left(self):
        """Accelerate player to the left."""
        self.facing_right = False
        if self.x_speed > -self.max_walk_speed:
            self.x_speed += -self.x_accel

    def go_right(self):
        """Accelerate player to the right."""
        self.facing_right = True
        if self.x_speed < self.max_walk_speed:
            self.x_speed += self.x_accel

    def stop(self):
        """Decelerate player on x axis."""
        if self.x_speed > 0:
            self.x_speed -= 1
        elif self.x_speed < 0:
            self.x_speed += 1

    def jump(self):
        """Move player contrary to gravity if on ground."""
        if self.is_touching_ground() or self.is_climbing():
            self.assign_default_physics()
            self.y_speed = -26
            # XXX
            # create a container to manage this
            # automatically and also verify
            # busy states to prevent underruns
            # errors (probably? need to know more)
            # The automation part is desired anyway,
            # since this kind of setup should be used
            # lots of times
            choice(self.jump_sounds).play()

    def use_equipped_item(self):
        """Request item to inventory and use if available."""
        item = self.inventory_manager.request_equipped_item()

        if item:

            item.prepare_for_usage(self)
            self.level.add_obj_to_group(item, "equippable_items")

    ### Interaction methods (usually refer other objects)

    def up_action(self):
        """Check for possible actions using up movement.

        Options available depending on conditions:
        - Enter a door/cave/entrance to another level.
        - Climb ladders
        """
        if self.is_touching_ladder():
            self.assign_climbing()
            if not self.is_leaving_ladder():
                self.y_speed = -self.climb_speed
        elif self.is_touching_portal():
            # XXX
            # After entrance animation is timed
            # use said time here before changing level
            # so user see the entrance animation
            # before 'entering'
            portal = self.portal_holder.pop()
            self.portal_holder = None
            after_milliseconds = 100
            distance = get_straight_distance(self.rect.midbottom, portal.rect.midbottom)

            # 38 is arbitrary, just means it's close.
            if distance <= 38 and self.health and self.is_touching_ground():
                set_level_switching(
                    after_milliseconds, portal.next_level, portal.destination
                )

    def down_action(self):
        """Check for possible actions using up movement.

        Options available depending on conditions:
        - Crouch
        - Climb ladders
        """
        if self.is_touching_ground() and not self.x_speed:
            if self.facing_right:
                self.anim_player.ensure_animation("crouched_idle")
            else:
                self.anim_player.ensure_animation("crouched_idle_reverse")

        elif self.is_touching_ladder() and self.ladder_to_ground_cooldown.finished:
            self.assign_climbing()
            self.y_speed = self.climb_speed

    def suffer_damage(self, damage):
        """Subtract damage from player healthbar."""
        # XXX
        # pc should scream lightly but convincingly
        # when taking damage, to inspire sympathy and
        # not scare the player.
        self.healthbar.update_width(-damage)

    # also controlled by user
    def interact(self):
        """Trigger interactions if available.

        This method detects interactive objects near the
        player and trigger whatever interaction they
        may have. Such interactions may include
        dialogues, monologues, getting items, etc."""
        actors_onscreen = self.level.onscreen_map["actors"]
        middle_props_onscreen = self.level.onscreen_map["middle_props"]
        props = []
        props.extend(actors_onscreen)
        props.extend(middle_props_onscreen)

        if props:
            # XXX
            # Change this to not just probe the closest
            # prop, but to keep probing them until it finds
            # one which has interactions (in case the first
            # don't have).
            sprite_distance_pairs = []
            for prop in props:
                distance = get_straight_distance(prop.rect.center, self.rect.center)
                if distance < 128:
                    sprite_distance_pairs.append((prop, distance))

            closer_prop_pair = min(
                sprite_distance_pairs, default=None, key=lambda seq: seq[1]
            )

            if closer_prop_pair:

                closer_prop = closer_prop_pair[0]

                try:
                    closer_prop.receive_interaction(self)
                except AttributeError as err:

                    if "receive_interaction" in str(err):
                        msg = "Expected AttributeError" + " was silenced."
                        logger.info(msg)

                    else:
                        msg = "Unexpected Attribute Error"
                        logger.exception(msg)
                        raise err

    def send_dialogue(self, dialogue):
        """Send dialogue to dialogue_manager."""
        self.dialogue_manager.get_dialogue(dialogue)

    def send_prompt(self, prompt):
        """Send dialogue to dialogue_manager."""
        self.prompt_manager.get_prompt(prompt)

    def confirm_prompt(self):
        """Confirm current prompt, if any."""
        self.prompt_manager.confirm()

    def finish_dialogue(self):
        """Skip entire dialogue."""
        self.dialogue_manager.finish_dialogue()

    def get_item(self, item, report_to=None):
        """Add item on inventory.

        item
            An instance of whichever class can be used as
            an item by player.
        report_to
            A reference to a instance of a class that
            needs to have the execution of this get_item
            method reported. If not provided, we'll
            report the execution to the item itself,
            so it can execute any administrative tasks
            as needed.
        """
        if self.inventory_manager.add(item):

            if report_to:
                report_to.got_item()
            else:
                item.got_item()

    def say_full_inventory(self):
        """Play a 'full inventory' dialogue."""
        self.send_dialogue(self.full_inventory_dialogue)

    ### Support methods

    def is_touching_ground(self):
        """Return True if touching ground."""
        self.rect.y += 2
        platform_bellow = self.level.onscreen_map["blocks"].collide_any(self)
        self.rect.y += -2

        if platform_bellow:
            return True
        else:
            return False

    def is_leaving_ladder(self):
        """Return True if leaving a ladder by going up."""
        self.rect.y += -self.climb_speed
        leaving_ladder = not self.is_touching_ladder()
        self.rect.y += self.climb_speed
        return leaving_ladder

    def is_touching_ladder(self):
        """Return True if touching a ladder."""
        touching_ladder = False

        middle_props = self.level.onscreen_map["middle_props"].collide(self)

        if middle_props:
            touching_ladder = [
                obj for obj in middle_props if "climbable" in obj.prop_name
            ]

        return touching_ladder

    def is_touching_portal(self):
        """Return True if touching a portal."""
        touching_portal = False

        middle_props = self.level.onscreen_map["middle_props"].collide(self)

        # XXX
        # use portal substring to indicate a portal
        # or use a better way if you think of any
        if middle_props:
            touching_portal = [
                obj
                for obj in middle_props
                if obj.prop_name in ("simple_door", "cave_entrance")
            ]
            self.portal_holder = touching_portal

        return touching_portal

    def is_climbing(self):
        """Return True if playing is climbing."""
        return self.movement_routine == self.climb_physics

    def is_crouched(self):
        """Return True if player is crouched."""
        return self.anim_player.anim_name in ("crouched_idle", "crouched_idle_reverse")

    def die(self):
        """Start a death animation and cue level restart.

        Also play a death SFX."""
        # XXX
        # 2018-01-26 06:58:57
        # Death animation trigger here in the future
        # Also adjust milliseconds to allow animation
        # to play through and display some text on screen
        # like "you died", etc.

        # XXX
        # 2018-01-26 06:59:30
        # Death SFX trigger here in the future
        print("Died")
        milliseconds = 2000
        set_timer(RESTART_FROM_SAVE, milliseconds)

    def stand_up(self):
        """Ensures player stand up stance if on ground."""
        if self.is_touching_ground():

            if self.x_speed:

                if self.facing_right:
                    self.anim_player.ensure_animation("walk")

                else:
                    self.anim_player.ensure_animation("walk_reverse")
            else:

                if self.facing_right:
                    self.anim_player.ensure_animation("idle")

                else:
                    self.anim_player.ensure_animation("idle_reverse")

    def skip_interactions(self):
        self.dialogue_manager.skip_dialoguebox()
        self.prompt_manager.finish_prompt()

    def map_actions(self):
        """Create map and pass to control overrider.

        This map may be used by the control overrider class
        to control player behaviour.
        """
        actions = {
            "advance_deny": self.skip_interactions,
            "interact": self.interact,
            "fire_item": self.use_equipped_item,
            "jump": self.jump,
            "player_left": self.go_left,
            "player_right": self.go_right,
            "stop": self.stop,
        }

        register_actions("player", actions)



"""Facility for interactive.main.Level class.

Interactive levels are instances of the Level class in this
module. They manage classes which are considered interactive
because the player character can interact with them directly.
This is the level used for the gameplay.
"""

from shutil import copyfile
from os.path import basename, dirname, isfile
from functools import partial, partialmethod

### third-party imports

from pygame import QUIT, KEYDOWN, K_m, K_KP0, K_KP1, K_ESCAPE

from pygame.color import THECOLORS
from pygame.time import set_timer
from pygame.key import get_pressed as get_pressed_keys
from pygame.event import get as get_events, clear as clear_events
from pygame.display import update

### local imports

from ..config import (
    GAME_REFS,
    SCREEN,
    SCREEN_WIDTH,
    SCREEN_HEIGHT,
    SWITCH_LEVEL_TRIGGER,
    RESTART_FROM_SAVE,
    KEYS_MAP,
)

from ..common.math import calculate_jump, unscroll_coordinates
from ..common.behaviour import CallList
from ..common.jsonhandler import load_json, save_json
from ..common.timehandler import get_isostring_datetime

from ..appcommon.task import add_task, update_task_manager, clear_task_manager
from ..appcommon.surf import render_image
from ..appcommon.dialog import clear_messages
from ..appcommon.autoblit import BlitterSet
from ..appcommon.exception import (
    QuitGameException,
    LevelSwitchException,
    RestartLevelException,
    ManagerSwitchException,
)
from ..appcommon.collision import get_objs_onscreen
from ..appcommon.overrider import prepare_script, perform_actions
from ..appcommon.loadscreen import blit_loading_screen

from .classes import get_instance, insert_level_reference, remove_level_reference

from ..player.main import Player

from ..fpsdisplay import fps_display

from .escapemenu import EscapeMenu
from .tabletmenu import TabletMenu

from .scroll import set_scroll_handles
from .groups import set_additional_groups

from ..music import MUSIC_MANAGER
from ..palette import BLACK


# XXX
# maybe in the future use a kind of 'funclist' object,
# a container of callabels whose __call__ method
# causes each callable to be executed, to group
# functionalities that are closely tied together
# like the event and input handling in this class.


class GameLevel(object):
    """A interactive level class."""

    def __init__(self, level_filepath, save_state_filepath, destination=None):
        """Set variables and perform setups.

        level_filepath
            The path of the current level being loaded.
        save_state_filepath
            Filepath for the state.stt json file, which
            contains information of the slot being used
            and other game states and player info/stats.
        destination (defaults to None)
            The bottomleft coordinates wherein to place the
            player when s/he enters a level via some
            kind of entrance. If None, the player will
            be either put on the center of the first screen
            or on the last spot where s/he saved the game.
        """
        self.level_filepath = level_filepath
        self.save_state_filepath = save_state_filepath

        self.level_name = basename(level_filepath)
        self.destination = destination

        self.perform_level_data_setup()
        self.initialize_level()

    # XXX explain why it is important to always use swap
    # files in the docstring below
    def perform_level_data_setup(self):
        """Prepare level and state data structures.

        Here we check for the existence of swap files for
        the level and state file. If they exist, we load
        them, if they don't, we create them. Whichever the
        case, we always work with swap files.
        """
        ### Handle level swap file

        self.level_swap_path = "{}.swp".format(self.level_filepath)

        if not isfile(self.level_swap_path):
            copyfile(self.level_filepath, self.level_swap_path)

        self.level_data = load_json(self.level_swap_path)

        ### Handle state swap file

        self.state_swap_path = "{}.swp".format(self.save_state_filepath)

        if not isfile(self.state_swap_path):
            copyfile(self.save_state_filepath, self.state_swap_path)

        self.save_state_data = load_json(self.state_swap_path)

    def initialize_level(self):
        """Set all variables and instantiate objects.

        This sets all needed data as well as populate
        the level with the needed objects."""
        ### Initial setup

        blit_loading_screen()

        self.horizontal_scroll = self.vertical_scroll = 0

        self.background = render_image("sky_background.png")

        ### Setting game objects grouping

        self.objs_to_scroll = BlitterSet()
        self.huds_group = BlitterSet()

        self.group_names = [
            "back_props",
            "middle_props",
            "actors",
            "equippable_items",
            "blocks",
            "front_props",
        ]

        self.group_map = {}
        self.onscreen_map = {}

        for group_name in self.group_names:

            self.group_map[group_name] = BlitterSet()
            self.onscreen_map[group_name] = BlitterSet()

        insert_level_reference(self)

        for group in self.group_map:
            self.set_group(group)

        set_additional_groups(self)

        ### Setting other objects

        self.set_escape_menu()
        self.set_tablet_menu()

        set_scroll_handles(self)

        self.fps_display = fps_display

        if not self.destination:
            self.scroll(
                self.level_data.get("horizontal_scroll", 0),
                self.level_data.get("vertical_scroll", 0),
            )

        self.set_player()

        MUSIC_MANAGER.cue(self.level_data.get("mood", "casual"))

        ### Store event handling method references

        self.normal_event_handling = CallList(
            [self.act_on_events, self.act_on_key_input]
        )

        ### Assign methods to action routines
        self.get_control()

    def set_group(self, group_name):
        """Configure props in group.

        group_name
            A string representing the name of the group being
            set. This name is defined in the body of the
            initialize_level method.
        """
        group = self.group_map[group_name]
        for obj_data in self.level_data.get(group_name, set()):
            obj = get_instance(obj_data)
            obj.json_data = obj_data
            group.add(obj)

        self.objs_to_scroll |= group

    def add_obj_to_group(self, obj, group_name, at_end=False):
        """Add obj in level relevant groups.

        obj
            Any gaming object.
        group_name
            The name of any gaming set group create in
            self.initialize_level method.
        at_end
            Boolean indicating if the operation should be
            done at the end of the loop or not. Default is
            False so the changes are made immediatelly.
            True should be passed when an object inside
            an specific group wants to add another one
            in the same group in order to prevent the
            group to be changed while we iterate over it.
        """
        if at_end:

            callables = (
                self.group_map[group_name].add,
                self.onscreen_map[group_name].add,
                self.objs_to_scroll.add,
            )

            for callable_task in callables:
                callable_partial = partial(callable_task, obj)
                add_task(callable_partial, 0)

        else:

            self.group_map[group_name].add(obj)
            self.onscreen_map[group_name].add(obj)
            self.objs_to_scroll.add(obj)

    def remove_obj_from_group(self, obj, group_name, at_end=False):
        """Remove obj in level relevant groups.

        obj
            Any gaming object.
        group_name
            The name of any gaming set group create in
            self.initialize_level method.
        at_end
            Boolean indicating if the operation should be
            done at the end of the loop or not. Default is
            False so the changes are made immediatelly.
            True should be passed when an object inside
            an specific group wants to add another one
            in the same group in order to prevent the
            group to be changed while we iterate over it.
        """
        if at_end:

            callables = (
                self.group_map[group_name].remove,
                self.onscreen_map[group_name].remove,
                self.objs_to_scroll.remove,
            )

            for callable_task in callables:
                callable_partial = partial(callable_task, obj)
                add_task(callable_partial, 0)

        else:
            self.group_map[group_name].remove(obj)
            self.onscreen_map[group_name].remove(obj)
            self.objs_to_scroll.remove(obj)

    def set_escape_menu(self):
        """Set an escape menu for system options."""
        self.escape_menu = EscapeMenu(self)

    def set_tablet_menu(self):
        """Set a tablet menu, a gameplay display object.

        The tablet menu as the name implies represents an
        electronic tablet and is as a convenience for
        players to remember gameplay data, check inventory and
        read past disclosed information, etc.
        It isn't absolutely needed in gameplay.
        """
        self.tablet_menu = TabletMenu(self)

    def set_player(self):
        """Configure player.

        The player might need to be placed in a more general
        or specific position, depending on whether
        destination coordinates were provided on the
        __init__ constructor.
        """
        if self.destination:

            for prop in self.group_map["middle_props"]:

                try:
                    if prop.portal_name == self.destination:
                        start_position = prop.rect.bottomleft

                except AttributeError:
                    pass

        else:

            start_position = self.level_data.get("player_position", (640, 360))

        health = self.save_state_data.get("health", 100)
        self.player = Player(
            unscroll_coordinates(start_position),
            level=self,
            state_data=self.save_state_data,
            health=health,
        )

        if self.destination:
            x_scrolling, y_scrolling = calculate_jump(start_position)
            self.scroll(x_scrolling, y_scrolling)

    def get_onscreen_data(self):
        """Separate sprites on the screen.

        This is done whenever the player 'enters' a new
        screen, that is, whenever s/he reaches the edges
        of the screen to enter a new one.

        By separating the sprites on the screen from the
        ones outside we optimize resources by only
        update sprites on the screen (since it doesn't
        make any difference to update sprites we don't
        see, well, in most cases at least).
        """
        for group in self.onscreen_map.values():
            group.clear()

        for group_name in self.group_map:
            group = self.group_map[group_name]
            onscreen_objs = get_objs_onscreen(group)
            self.onscreen_map[group_name] |= onscreen_objs

    def give_up_control(self):
        """Give control to overrider."""
        prepare_script()
        self.control = perform_actions

    def get_control(self):
        """Setup control routines.

        Used when constructing the instance and also when
        getting control back from overrider.
        """
        clear_events()
        self.control = self.normal_event_handling

    def update(self):
        """Update objects in level instance.

        When an instance of this class is the update manager,
        this method will be called. It is responsible for
        updating level objects. However, only those on the
        screen are update in reality to save resources."""
        for group_name in self.group_names:
            self.onscreen_map[group_name].update_objs()
            if group_name == "actors":
                self.player.update()

        self.huds_group.update_objs()

        update_task_manager()

    def draw(self):
        """Draw objects."""
        SCREEN.blit(self.background, (0, 0))

        for group_name in self.group_names:
            self.onscreen_map[group_name].draw_objs()
            if group_name == "actors":
                self.player.draw()

        self.huds_group.draw_objs()

        update()

    def act_on_events(self):
        """Perform actions based on events."""
        for event in get_events():

            if event.type == QUIT:
                raise QuitGameException

            elif event.type == KEYDOWN:

                if event.key == K_ESCAPE:
                    raise ManagerSwitchException(self.escape_menu)

                elif event.key == KEYS_MAP["advance_deny"]:
                    self.player.skip_interactions()

                elif event.key == KEYS_MAP["accept_prompt"]:
                    self.player.confirm_prompt()

                elif event.key == K_KP0:

                    if self.fps_display in self.huds_group:
                        self.huds_group.remove(self.fps_display)
                    else:
                        self.huds_group.add(self.fps_display)

                elif event.key == K_KP1:

                    if self.scroll_barriers.issubset(self.huds_group):

                        for item in self.scroll_barriers:
                            self.huds_group.remove(item)

                    else:
                        self.huds_group.update(self.scroll_barriers)

                elif event.key == KEYS_MAP["interact"]:
                    self.player.interact()

                elif event.key == KEYS_MAP["fire_item"]:
                    self.player.use_equipped_item()
                # XXX
                # this works, but let's omit it's usage
                # by now.
                # elif event.key == K_m:
                #    raise ManagerSwitchException(
                #                           self.tablet_menu)

                elif event.key == KEYS_MAP["jump"]:
                    self.player.jump()

            elif event.type == SWITCH_LEVEL_TRIGGER:
                self.switch_level()

            elif event.type == RESTART_FROM_SAVE:
                set_timer(RESTART_FROM_SAVE, 0)
                raise RestartLevelException

    def act_on_key_input(self):
        """Perform actions based on key input."""
        key_input = get_pressed_keys()

        player_left, player_right = (
            key_input[KEYS_MAP["player_left"]],
            key_input[KEYS_MAP["player_right"]],
        )

        player_up = key_input[KEYS_MAP["player_up"]]
        player_down = key_input[KEYS_MAP["player_down"]]

        if player_left and not player_right:
            self.player.go_left()

        elif player_right and not player_left:
            self.player.go_right()

        else:
            self.player.stop()

        if player_up and not player_down:
            self.player.up_action()

        elif player_down and not player_up:
            self.player.down_action()

        elif not player_up and not player_down:
            self.player.stand_up()

    def scroll(self, dx, dy):
        """Scroll all objects in x and y amounts.

        dx, dy
            Integers representing scrolling in x and y
            axes, respectively.
        """
        try:
            self.player.finish_dialogue()

        except AttributeError:
            pass

        self.horizontal_scroll += dx
        self.vertical_scroll += dy

        self.objs_to_scroll.scroll(dx, dy)
        self.get_onscreen_data()

    scroll_left = partialmethod(scroll, SCREEN_WIDTH, 0)
    scroll_right = partialmethod(scroll, -SCREEN_WIDTH, 0)
    scroll_up = partialmethod(scroll, 0, SCREEN_HEIGHT)
    scroll_down = partialmethod(scroll, 0, -SCREEN_HEIGHT)

    def switch_level(self):
        """Switch to next level."""
        set_timer(SWITCH_LEVEL_TRIGGER, 0)

        clear_task_manager()
        clear_messages()

        self.save_states_buffer()
        remove_level_reference()

        level_name = GAME_REFS.next_level
        del GAME_REFS.next_level

        raise LevelSwitchException(level_name)

    def pre_saving_routine(self):
        """Execute pre-saving routines on game objects."""
        for obj in self.objs_to_scroll:

            try:
                obj.pre_saving_routine()

            except AttributeError:
                pass

    def save_states_buffer(self):
        """Save level and game states on buffer."""
        ### Level states

        self.level_data["player_position"] = self.player.rect.bottomleft
        self.level_data["horizontal_scroll"] = self.horizontal_scroll
        self.level_data["vertical_scroll"] = self.vertical_scroll

        save_json(self.level_data, self.level_swap_path)

        ### Game states

        self.save_state_data["last_level"] = self.level_name
        self.save_state_data["last_played_datetime"] = get_isostring_datetime()
        self.save_state_data["health"] = self.player.health

        save_json(self.save_state_data, self.state_swap_path)



"""Configuration file."""

from os import environ

from types import SimpleNamespace
from pathlib import Path

### third-party imports

from pygame import USEREVENT

from pygame.display import set_mode, set_icon, set_caption
from pygame.image import load as load_image
from pygame.time import Clock

### local imports

from .common.jsonhandler import save_json, load_json

from .palette import ARBITRARY_COLORKEY
from .keymapping import DEFAULT_KEYS_MAP


### locations

data_dir = Path(__file__).parent / "data"

MUSIC_DIR = str(data_dir / "music")
LEVELS_DIR = str(data_dir / "levels")
SOUNDS_DIR = str(data_dir / "sounds")
SCENES_DIR = str(data_dir / "scenes")
ANIMATIONS_DIR = str(data_dir / "animations")

IMAGES_DIR = str(data_dir / "images")
IMAGES_META_FILE = str(data_dir / "images_meta_file.json")

FONTS_DIR = str(data_dir / "fonts")
FONTS_META_FILE = str(data_dir / "fonts_meta_file.json")

### defining path to config file

if "APPDATA" in environ:
    config_dir = Path(environ["APPDATA"])

elif "XDG_CONFIG_HOME" in environ:
    config_dir = Path(environ["XDG_CONFIG_HOME"])

else:
    config_dir = Path(environ["HOME"]) / ".config"


APP_CONFIG_DIR = config_dir / "danittr"

if not APP_CONFIG_DIR.exists():
    APP_CONFIG_DIR.mkdir(parents=True)

## create a "saves" directory if it doesn't exist

SAVES_DIR = APP_CONFIG_DIR / "saves"

if not SAVES_DIR.exists():
    SAVES_DIR.mkdir(parents=True)

SAVES_DIR = str(SAVES_DIR)

### Gamewide constants

## object to carry references to important data throughout
## the game run
GAME_REFS = SimpleNamespace()

## time related objects/constants

CLOCK = Clock()

FPS = 30
MILLISECS_PER_FRAME = (1 / FPS) * 1000

### Custom events

SWITCH_LEVEL_TRIGGER = USEREVENT + 1
RESTART_FROM_SAVE = USEREVENT + 2

### Other data

STATE_FILE_NAME = "state.stt"
DEFAULT_STATE_DATA = {
    "last_level": "level01.lvl",
    "last_played_datetime": None,
    "health": 100,
}

### Custom settings

CONFIG_FILE = str(APP_CONFIG_DIR / "config.json")

DEFAULT_SETTINGS = {
    "music_volume": 0.2,
    "sound_volume": 0.2,
    "keys_map": DEFAULT_KEYS_MAP,
}


try:
    USER_SETTINGS = load_json(CONFIG_FILE)
except Exception as err:
    print(str(err))
    USER_SETTINGS = DEFAULT_SETTINGS
    save_json(USER_SETTINGS, CONFIG_FILE)

KEYS_MAP = USER_SETTINGS["keys_map"]

### pygame setups and constants

## Set caption and icon for window
set_caption("Dani to The Rescue", "DtTR")
ICON = load_image(str(data_dir / "game_icon.png"))
ICON.set_colorkey(ARBITRARY_COLORKEY)
set_icon(ICON)

## Set the screen and screen rect

# XXX must load fullscreen setting

### Defining constants and performing screen setup (set_mode)
SCREEN_WIDTH, SCREEN_HEIGHT = 1280, 720
FLAG = 0
DEPTH = 32

SCREEN = set_mode((SCREEN_WIDTH, SCREEN_HEIGHT), FLAG, DEPTH)

SCREEN_RECT = SCREEN.get_rect()



"""Tools to improve styles."""

from pygame.draw import line as draw_line

from ..palette import HIGHLIGHT, SHADOW, BLACK


def give_depth_finish(surf, highlight_color=HIGHLIGHT, shadow_color=SHADOW):
    """Give a subtle finish which conveys depth.

    surf
        Instance of pygame.Surface.
    highlight_color
    shadow_color
        Tuple containing three integers representing
        red, green and blue values from 0 to 255.
        Used for highlights and shadows respectively.
    """
    width, height = surf.get_size()

    start = (0, 0)
    end = (width, 0)
    draw_line(surf, highlight_color, start, end, 1)

    start = (0, 0)
    end = (0, height)
    draw_line(surf, highlight_color, start, end, 1)

    start = (0, height - 1)
    end = (width - 1, height - 1)
    draw_line(surf, shadow_color, start, end, 1)

    start = (width - 1, 0)
    end = (width - 1, height - 1)
    draw_line(surf, shadow_color, start, end, 1)

    return surf



"""Facility for resource loading.

Functions and tools to easily create objects
and surfaces.
"""
from os.path import join

### third-party imports

from pygame import Surface
from pygame.image import load as load_image
from pygame.display import update

### local imports

from ..config import IMAGES_DIR, IMAGES_META_FILE

from ..common.jsonhandler import load_json
from .autoblit import BasicObject

from ..palette import BLACK


INVISIBLE_SURF = Surface((0, 0)).convert()

IMAGES_MAP = {}

IMAGE_NAME_TO_ALPHA_MAP = load_json(IMAGES_META_FILE)

### functions for general usage


def render_image(
    img_name, return_obj=False, coordinates_name="topleft", coordinates_value=(0, 0)
):
    """Return a surface or object representing an image.

    img_name (string)
        The basename of an image to be loaded.
    return_obj (boolean)
        Boolean indicating whether or not a
        appcommon.autoblit.BasicObject instance should be
        returned. Default is False, and causes a
        pygame.Surface instance to be returned instead.
        In case this parameter is True, the coordinates_name
        and coordinates_value parameters are used.
    coordinates_name
        A string specifying which of the rect attributes
        should be used to assign the coordinates value.
    coordinates_value
        A 2-tuple of integers representing a point to be
        assigned to one of the rects coordinates.
    """
    try:
        surf = IMAGES_MAP[img_name]

    except KeyError:

        ### join dir and image name and convert to string to
        ### form path
        img_path = join(IMAGES_DIR, img_name)

        ### load surface using appropriate method according
        ### to presence of per pixel alphas
        if IMAGE_NAME_TO_ALPHA_MAP[img_name]:
            surf_method = Surface.convert_alpha

        else:
            surf_method = Surface.convert

        surf = surf_method(load_image(img_path))

        IMAGES_MAP[img_name] = surf

    ### return either the surface or a customized object
    ### containing it

    if return_obj:

        ## instantiate object and set its attributes

        obj = BasicObject()
        obj.image = surf
        obj.rect = obj.image.get_rect()

        ## position its rect
        setattr(obj.rect, coordinates_name, coordinates_value)

        return obj

    else:
        return surf


def render_rect(
    width,
    height,
    color=BLACK,
    return_obj=False,
    coordinates_name="topleft",
    coordinates_value=(0, 0),
):
    """Return surface or object representing a filled rect.

    width
        Int indicating the desired width of the generated
        surface.
    height
        Int indicating the desired height of the generated
        surface.
    color
        A tuple or list of r, g, b values which are
        integers ranging from 0 to 255. Used to fill the
        generated surface. An optional fourth value can be
        passed, which is an integer in the same range as the
        others, representing the image opacity (0 for full
        transparency and 255 for full opacity).
    return_obj
        Boolean indicating whether or not a
        appcommon.autoblit.BasicObject instance should be
        returned. Default is False, and causes a
        pygame.Surface instance to be returned instead.
        In case this parameter is True, the coordinates_name
        and coordinates_value parameters are used.
    coordinates_name
        A string specifying which of the rect attributes
        should be used to assign the coordinates value.
    coordinates_value
        A 2-tuple of integers representing a point to be
        assigned to one of the rects coordinates.
    """
    ### instantiate surface taking transparency into account

    try:
        has_transparency = color[3] < 255
    except IndexError:
        has_transparency = False
    finally:

        if has_transparency:
            surf = Surface((width, height)).convert_alpha()

        else:
            surf = Surface((width, height)).convert()

    surf.fill(color)

    ### return either the surface or a customized object
    ### containing it

    if return_obj:

        ## instantiate object and set its attributes

        obj = BasicObject()
        obj.image = surf
        obj.rect = obj.image.get_rect()

        ## position its rect
        setattr(obj.rect, coordinates_name, coordinates_value)

        return obj

    else:
        return surf


# XXX maybe raise a ValueError if at least one dimension
# of the original surf is larger than the required
# dimensions? Ponder.
def enlarge_surf(surf, width, height, background_color=(*BLACK, 0)):
    """Blit current surf on bigger one to enlarge it.

    width
        Int indicating the final desired width of the
        generated surface.
    height
        Int indicating the final desired height of the
        generated surface.
    background_color
        A tuple or list of r, g, b values which are
        integers ranging from 0 to 255. Used to fill the
        generated surface.
    """
    ### instantiate surface taking transparency into account

    ## alias background color
    bg_color = background_color

    try:
        has_transparency = bg_color[3] < 255

    except IndexError:
        has_transparency = False

    finally:

        if has_transparency:
            bg_surf = Surface((width, height)).convert_alpha()

        else:
            bg_surf = Surface((width, height)).convert()

    bg_surf.fill(bg_color)

    bg_surf.blit(surf, (0, 0))

    return bg_surf


## XXX
## About using colorkey:
## Ideally, all images that need transparency
## would use colorkey, but though its technically
## good, it is artistically/aesthetically bad
## for what I want to achieve, because edges
## become jagged/serrated. Maybe I'll find a
## workaround in the future.



"""Facility for showing temporary messages."""

### third-party imports

from pygame import QUIT, KEYDOWN

from pygame.display import update
from pygame.event import get as get_events

### local imports

from ..config import SCREEN, SCREEN_RECT

from ..palette import WHITE

from ..common.behaviour import empty_function

from .task import update_task_manager
from .text.main import render_text
from .exception import QuitGameException, ManagerSwitchException
from .surf import render_rect


class MessageScreen(object):
    """A temporary update manager to show messages."""

    def __init__(self):
        """Assign variables and perform setups."""
        self.message_map = {}
        self.update_manager = None
        self.current_message = None
        self.extra_routine = empty_function

        self.background = render_rect(*SCREEN_RECT.size)
        self.set_press_prompt()

    def set_press_prompt(self):
        """Set a surface and coordinates for prompt message.

        Such message is a prompt for the user to press
        any key in order to remove the message screen.
        """
        self.press_prompt = render_text(
            "Press any key to exit",
            font_size=32,
            foreground_color=WHITE,
            return_obj=True,
        )

        self.press_prompt.rect.topright = SCREEN_RECT.topright

        self.press_prompt.rect.x += -100
        self.press_prompt.rect.y += 35

    def cache_message(self, message_name, text):
        """Create and store a text obj for later use.

        message_name
            String representing a field in a dict which
            is used to retrieve information later.
        text
            String used to generated text surface.
        """
        obj = render_text(text, font_size=32, foreground_color=WHITE, return_obj=True)

        obj.rect.center = SCREEN_RECT.center

        self.message_map[message_name] = obj

    def display_message(self, message_name, update_manager, extra_routine=None):
        """Retrieve from map and display surface.

        message_name
            String representing the field in a dict whose
            value is the surface to be displayed.
        update_manager
            Any update_manager instance which
            should be reassigned back to
            game.Game.update_manager back after
            showing messages.
        extra_routine
            Optional callable to be executed after
            displaying the message.
        """
        if extra_routine:
            self.extra_routine = extra_routine

        self.update_manager = update_manager
        self.current_message = message_name

        raise ManagerSwitchException(self)

    def control(self):
        """Handle events."""
        for event in get_events():

            if event.type == QUIT:
                raise QuitGameException

            elif event.type == KEYDOWN:
                self.exit_message_screen()

    def update(self):
        """Blit message on screen."""
        self.extra_routine()
        update_task_manager()

    def draw(self):
        """Draw objects."""
        SCREEN.blit(self.background, (0, 0))

        message = self.message_map[self.current_message]
        SCREEN.blit(message.image, message.rect)

        SCREEN.blit(self.press_prompt.image, self.press_prompt.rect)

        update()

    def exit_message_screen(self):
        """Get back to previous update_manager."""
        raise ManagerSwitchException(self.update_manager)

    def clear(self):
        """Clear messages and other data."""
        self.message_map.clear()
        self.update_manager = None
        self.current_message = None
        self.extra_routine = empty_function


MESSAGE_SCREEN = MessageScreen()
cache_message = MESSAGE_SCREEN.cache_message
display_message = MESSAGE_SCREEN.display_message
clear_messages = MESSAGE_SCREEN.clear



"""Facility for the NPC (non-playable character) class."""

from ..appcommon.autoblit import BasicObject
from ..appcommon.dialoguefactory import dialoguebox_factory
from ..common.math import unscroll_coordinates
from ..common.behaviour import empty_function
from ..appcommon.surf import render_image


class NPC(BasicObject):
    """A non-playable character."""

    level = None

    prop_map = {"npc": render_image("npc.png")}

    def __init__(self, prop_name, coordinates, dialogue=None):
        """Assign variables."""
        self.prop_name = prop_name

        self.image = NPC.prop_map[prop_name]

        self.rect = self.image.get_rect()

        self.rect.bottomleft = coordinates

        if dialogue:
            object_screen_midtop = unscroll_coordinates(self.rect.midtop)
            offset_screen_midtop = (
                object_screen_midtop[0],
                object_screen_midtop[1] - 32,
            )
            self.dialogue = dialoguebox_factory(dialogue.copy(), offset_screen_midtop)

        else:
            self.dialogue = []

        self.update = empty_function

    def receive_interaction(self, player):
        """Send dialogue to player state manager."""
        player.send_dialogue(self.dialogue)

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)
